import pandas as pd
import numpy as np
import os
from PIL import ImageColor

from treelib import Tree

from vedo import load, merge, write

import brainrender
from brainrender.Utils.paths_manager import Paths
from brainrender.Utils.webqueries import request
from brainrender.Utils.data_io import load_mesh_from_file

"""
    Class to download and render a number of insect brains from:
    https://insectbraindb.org/app/

    The InsectBrainDB.org is primarily curated by Stanley Heinze

    The insectbraindb.org has a terms of use, which provides guidance on how best to credit data from these repositories. 

    The insectbraindb.org is primarily curated by Dr. Stanley Heinze, and 
    was buily by Kevin Tedore, and has several significant supporters, including the ERC.




    Check Examples/custom_atlases/insects_brains.py for more details.

"""


class IBDB(Paths):

    atlas_name = "InsectBrains"
    mesh_format = "obj"

    _base_url = "https://insectbraindb.org"
    _url_paths = dict(
        brain_info="archive/species/most_current_permitted/?species_id=",
        species_info="api/species/min/",
        data="https://s3.eu-central-1.amazonaws.com/ibdb-file-storage",
        structures_tree="https://insectbraindb.org/api/structures/list_hierarchy",
    )

    default_camera = dict(
        position=[1035.862, 420.237, -3058.244],
        focal=[1253.28, 759.97, 428.418],
        viewup=[0.03, -0.995, 0.095],
        distance=3509.914,
        clipping=[2616.137, 4643.461],
    )

    def __init__(
        self, species=None, sex=None, base_dir=None, make_root=True, **kwargs
    ):
        self.make_root = make_root

        Paths.__init__(self, base_dir=base_dir, **kwargs)

        # Get a list of available species
        self.species_info = pd.DataFrame(
            request(
                f"{self._base_url}/{self._url_paths['species_info']}"
            ).json()
        )
        self.species = list(self.species_info.scientific_name.values)

        # Get selected species
        self.structures, self.region_names, self.region_acronyms = (
            None,
            None,
            None,
        )
        self.sel_species = species
        self.sex = sex
        self.get_brain(species=species, sex=sex)

    def get_brain_id_from_species_name(self, sel_species):
        if sel_species not in self.species:
            raise ValueError(
                f"The species {sel_species} is not among the available species {self.species}"
            )

        try:
            brain_id = self.species_info.loc[
                self.species_info.scientific_name == sel_species
            ]["id"].values[0]
        except:
            raise ValueError(
                f"Could not find brain data for species {sel_species}\n"
                + f"available species:\n{self.species_info}"
            )
        return brain_id

    def get_structures_hierarchy(self):
        def add_descendants_to_tree(tree, structure, parent_id=None):
            """
                Recursively goes through all the the descendants of a region and adds them to the tree
            """
            if parent_id is not None:
                tree.create_node(
                    tag=structure["name"],
                    identifier=structure["id"],
                    parent=parent_id,
                )
            else:
                tree.create_node(
                    tag=structure["name"], identifier=structure["id"],
                )

            if "children" not in structure.keys():
                return
            if structure["children"]:
                for child in structure["children"]:
                    add_descendants_to_tree(tree, child, structure["id"])

        structures_hierarchy = request(
            self._url_paths["structures_tree"]
        ).json()

        tree = Tree()
        tree.create_node(
            tag="root", identifier=0,
        )
        for supercategory in structures_hierarchy:
            add_descendants_to_tree(tree, supercategory, 0)

        self.structures_hierarchy = tree

    def get_structures_reconstructions(self, species, sex):
        iid = self.get_brain_id_from_species_name(species)

        data = request(
            f"{self._base_url}/{self._url_paths['brain_info']}{iid}"
        ).json()
        if not len(data):
            print("Could not get any data for this brain")
            return None

        reconstructions = [r["viewer_files"] for r in data["reconstructions"]]
        if not reconstructions:
            print(f"No data was found for {species}")
            return

        n_elems = [len(rec) for rec in reconstructions]

        if sex is None:
            try:
                sex = np.argmax(n_elems)
            except:
                raise ValueError("No data retrieved")

        if not n_elems[sex]:
            raise ValueError(f"No reconstructions found for {sex} {species}")
        else:
            reconstruction = reconstructions[sex]

        # Get data about the brain regions
        structures = dict(
            name=[],
            acronym=[],
            color=[],
            obj_path=[],
            hemisphere=[],
            parent=[],
            children=[],
        )

        for d in reconstruction:
            if "structures" not in d.keys():
                continue
            if not d["structures"]:
                continue

            hemi = d["structures"][0]["hemisphere"]
            if hemi is None:
                hemi = "both"

            if hemi == "right":
                name = d["structures"][0]["structure"]["name"] + "_R"
            elif hemi == "left":
                name = d["structures"][0]["structure"]["name"] + "_L"
            else:
                name = d["structures"][0]["structure"]["name"]

            abbr = d["structures"][0]["structure"]["abbreviation"]
            acro = (
                abbr + d["p_file"]["file_name"].split(abbr)[-1].split(".")[0]
            )

            if "_left" in acro:
                acro = acro.split("_left")[0] + "_left"
            elif "_right" in acro:
                acro = acro.split("_right")[0] + "_right"

            structures["obj_path"].append(d["p_file"]["path"])
            structures["hemisphere"].append(hemi.lower())

            structures["name"].append(name)
            structures["acronym"].append(acro)
            structures["color"].append(
                d["structures"][0]["structure"]["color"]
            )

            structures["parent"].append(
                d["structures"][0]["structure"]["parent"]
            )
            structures["children"].append(
                d["structures"][0]["structure"]["children"]
            )

        self.structures = pd.DataFrame(structures)

    def get_brain(self, species=None, sex=None):
        # Get metadata about the brain from database
        if species is None:
            species = self.sel_species

        if species is None:
            print("No species name passed")
            return None

        self.species = species
        self.sex = sex

        # Get reconstructions and metadata
        self.get_structures_hierarchy()
        self.get_structures_reconstructions(species, sex)

        # Prep some vars
        self.region_acronyms = list(self.structures["acronym"])
        self.regions = list(self.structures["name"])

        meshes_fld = os.path.join(self.ibdb_meshes_folder, self.species)
        if not os.path.isdir(meshes_fld):
            print("Creating folder at: {}".format(meshes_fld))
            os.makedirs(meshes_fld)
        self.meshes_folder = meshes_fld

        if self.make_root:
            self.make_root_mesh()

    def download_and_write_mesh(self, acronym, obj_path):
        print(f"Downloading mesh data for {acronym}")
        path = self.structures.loc[
            self.structures.acronym == acronym
        ].obj_path.values[0]
        url = f"{self._url_paths['data']}/{path}"

        # download and write .obj
        mesh_data = request(url).content.decode("utf-8").split("\n")
        with open(obj_path, "w") as f:
            for md in mesh_data:
                f.write(f"{md}\n")
            f.close()

        # return the vtk actor
        return load(obj_path)

    def make_root_mesh(self):
        if self.structures is None:
            return

        obj_path = os.path.join(self.meshes_folder, "root.vtk")
        if os.path.isfile(obj_path):
            return

        # Get the mesh for each brain region to create root
        meshes = [
            self._get_structure_mesh(reg) for reg in self.region_acronyms
        ]
        root = merge(meshes)
        write(root, obj_path)

    # ---------------------------------------------------------------------------- #
    #                          Replace base atlas methods                          #
    # ---------------------------------------------------------------------------- #

    def _get_structure_mesh(self, acronym, **kwargs):
        """
        Fetches the mesh for a brain region from the atlas.

        :param acronym: string, acronym of brain region
        :param **kwargs:

        """
        if self.structures is None:
            print("No atlas was loaded, use self.get_brain first")
            return None

        if acronym not in self.region_acronyms and acronym != "root":
            raise ValueError(
                f"Acronym {acronym} not in available regions: {self.structures}"
            )

        # Get obj file path
        if acronym != "root":
            mesh_format = self.mesh_format
        else:
            mesh_format = "vtk"
        obj_path = os.path.join(
            self.meshes_folder, "{}.{}".format(acronym, mesh_format)
        )

        # Load
        if self._check_obj_file(acronym, obj_path):
            mesh = load_mesh_from_file(obj_path, **kwargs)
            return mesh
        else:
            mesh = self.download_and_write_mesh(acronym, obj_path)
            return None

    def _check_valid_region_arg(self, region):
        """
        Check that the string passed is a valid brain region name.
        """
        if self.structures is None:
            return False
        else:
            if region not in self.region_acronyms:
                return False
        return True

    def _check_obj_file(self, region, obj_file):
        """
        If the .obj file for a brain region hasn't been downloaded already, this function downloads it and writes it.

        :param region: string, acronym of brain region
        :param obj_file: path to .obj file to write downloaded data.

        """
        # checks if the obj file has been downloaded already, if not it takes care of downloading it
        if not os.path.isfile(obj_file):
            self.download_and_write_mesh(region, obj_file)
        return True

    def get_region_color(self, regions):
        """
        Gets the RGB color of a brain region from the atlas.

        :param regions:  list of regions acronyms.

        """
        if not isinstance(regions, list):
            if not self._check_valid_region_arg(regions):
                return None
            return ImageColor.getrgb(
                self.structures.loc[
                    self.structures.acronym == regions
                ].color.values[0]
            )
        else:
            colors = []
            for region in regions:
                if not self._check_valid_region_arg(region):
                    return None
                colors.append(
                    ImageColor.getrgb(
                        self.structures.loc[
                            self.structures.acronym == region
                        ].color.values[0]
                    )
                )
            return colors

    def get_brain_regions(
        self,
        brain_regions,
        alpha=None,
        colors=None,
        use_original_color=True,
        **kwargs,
    ):
        if alpha is None:
            alpha = brainrender.DEFAULT_STRUCTURE_ALPHA

        if not isinstance(brain_regions, (list, tuple)):
            brain_regions = [brain_regions]

        # check the colors input is correct
        if not use_original_color:
            if colors is None:
                colors = [
                    brainrender.DEFAULT_STRUCTURE_COLOR
                    for reg in brain_regions
                ]
            else:
                if isinstance(colors, (list, tuple)):
                    if len(colors) != len(brain_regions):
                        raise ValueError("Wrong number of colors")
                else:
                    colors = [colors for reg in brain_regions]
        else:
            colors = [self.get_region_color(reg) for reg in brain_regions]

        actors = {}
        for region, color in zip(brain_regions, colors):
            actors[region] = self._get_structure_mesh(region, c=color)

        return actors



""" 
    This tutorial shows how to show the position of labelled cells in brainrender
"""
import pandas as pd

import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene
from brainrender.Utils.scene_utils import get_n_random_points_in_region

# Create a scene
scene = Scene(
    title="labelled cells"
)  # specify that you want a view from the top


# Gerate the coordinates of N cells across 3 regions
regions = ["MOs", "VISp", "ZI"]
N = 1000  # getting 1k cells per region, but brainrender can deal with >1M cells easily.

# Render regions
scene.add_brain_regions(regions, alpha=0.2)

# Get fake cell coordinates
cells = []  # to store x,y,z coordinates
for region in regions:
    region_cells = get_n_random_points_in_region(
        scene.atlas, region=region, N=N
    )
    cells.extend(region_cells)
x, y, z = [c[0] for c in cells], [c[1] for c in cells], [c[2] for c in cells]
cells = pd.DataFrame(
    dict(x=x, y=y, z=z)
)  # ! <- coordinates should be stored as a pandas dataframe


# Add cells
scene.add_cells(cells, color="darkseagreen", res=12, radius=25)

# render
scene.render()



"""
    This example shows how to create a scene that has a cartoony look 
    (good for schematics and illustrations)
"""

import brainrender

brainrender.SHADER_STYLE = "cartoon"  # gives actors a flat shading
from brainrender.scene import Scene

scene = Scene(title="cartoon look")
th = scene.add_brain_regions("TH", alpha=0.5)

# Create a black line around each actor
scene.add_silhouette(scene.root)
scene.add_silhouette(th, lw=3)

scene.render()



try:
    import cv2
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "You need opencv to save videos in brainrender, please install opencv with: "
        + "pip install opencv-python"
    )


from vedo import Video as VtkVideo
import datetime
import os

from brainrender.Utils.video import save_videocap_to_video


class Video(VtkVideo):
    # Redifine vedo.Video close method
    def __init__(self, *args, fmt="mp4", **kwargs):
        super().__init__(*args, **kwargs)
        self.format = fmt

    def get_cap_from_images_folder(self, img_format="%1d.png"):
        """
            It creates a cv2 VideoCaptur 'cap' from a folder of images (frames)
        """
        if not os.path.isdir(self.tmp_dir.name):
            raise ValueError(f"Folder {self.tmp_dir.name} doesn't exist")
        if not os.listdir(self.tmp_dir.name):
            raise ValueError(f"Folder {self.tmp_dir.name} is empty")

        # Create video capture
        cap = cv2.VideoCapture(os.path.join(self.tmp_dir.name, img_format))

        # Check all went well
        ret, frame = cap.read()
        if not ret:
            raise ValueError(
                "Something went wrong, can't read form folder: "
                + self.tmp_dir.name
            )
        else:
            cap.set(1, 0)  # reset cap to first frame
        return cap

    def close(self):
        """
            Takes a folder full of frames saved as images and converts it into a video.
        """
        # Save frames as video
        cap = self.get_cap_from_images_folder()
        save_videocap_to_video(
            cap,
            self.name + "." + self.format,
            self.format,
            int(self.fps),
            iscolor=True,
        )
        print(f"Saved video as: {self.name}.{self.format}")

        # Clean up
        self.tmp_dir.cleanup()


class BasicVideoMaker:
    """
        Wrapper around vedo Video class to facilitate the creation of videos from
        brainrender scenes.

        Use kwargs to specify:
            - save_fld: folder where to save video
            - save_name: video name
            - video_format: e.g. mp4
            - duration: video duration in seconds
            - niters: number of iterations (frames) when creating the video
            - fps: framerate of video
    """

    def __init__(self, scene, **kwargs):
        self.scene = scene

        # Parss keyword argumets
        self.save_fld = kwargs.pop("save_fld", self.scene.atlas.output_videos)
        self.save_name = kwargs.pop(
            "save_name",
            "brainrender_video_"
            + f'_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}',
        )
        self.video_format = kwargs.pop("video_format", "mp4")

        self.duration = kwargs.pop("duration", 3)
        self.niters = kwargs.pop("niters", 60)
        self.fps = kwargs.pop("fps", 30)

    def parse_kwargs(self, **kwargs):
        """
            Parses arguments for video creation
            Use kwargs to specify:
                - save_fld: folder where to save video
                - save_name: video name
                - video_format: e.g. mp4
                - duration: video duration in seconds
                - niters: number of iterations (frames) when creating the video
                - fps: framerate of video

            Arguments not specified in kwargs will be assigned default values
        """
        self.save_fld = kwargs.pop("save_fld", self.save_fld)
        self.save_name = kwargs.pop("save_name", self.save_name)
        self.video_format = kwargs.pop("video_format", self.video_format)
        self.duration = kwargs.pop("duration", None)
        self.niters = kwargs.pop("niters", self.niters)
        self.fps = kwargs.pop("fps", self.fps)

    def make_video(self, azimuth=0, elevation=0, roll=0, **kwargs):
        """
        Creates a video using user defined parameters

        :param azimuth: integer, specify the rotation in degrees per frame on the relative axis. (Default value = 0)
        :param elevation: integer, specify the rotation in degrees per frame on the relative axis. (Default value = 0)
        :param roll: integer, specify the rotation in degrees per frame on the relative axis. (Default value = 0)
        :param kwargs: use to change destination folder, video name, fps, duration ... check 'self.parse_kwargs' for details. 
        """
        self.parse_kwargs(**kwargs)

        curdir = (
            os.getcwd()
        )  # we need to cd to the folder where the video is saved and then back here
        os.chdir(self.save_fld)
        print(f"Saving video in {self.save_fld}")

        # Create video
        video = Video(
            name=self.save_name,
            duration=self.duration,
            fps=self.fps,
            fmt=self.video_format,
        )

        # Render the scene first
        self.scene.render(interactive=False)

        # Make frames
        for i in range(self.niters):
            self.scene.plotter.show()  # render(interactive=False, video=True)  # render the scene first
            self.scene.plotter.camera.Elevation(elevation)
            self.scene.plotter.camera.Azimuth(azimuth)
            self.scene.plotter.camera.Roll(roll)
            video.addFrame()

        self.scene.close()
        video.close()  # merge all the recorded frames

        # Cd bake to original dir
        os.chdir(curdir)


class CustomVideoMaker(BasicVideoMaker):
    """
        Subclasses BasicVideoMaker and replaces make_video method.
    """

    def __init__(self, scene, **kwargs):
        BasicVideoMaker.__init__(self, scene, **kwargs)

    def make_video(self, video_function, **kwargs):
        """
        Let's users use a custom function to create the video.
        The custom function must:
            - have a 'scene' keyword argument to accept a Scene() instance
            - have a 'videomaker' keyword argument to accept the CustomVideoMaker (self) instance
            - have a 'video' keyword that takes the Video argument
            - return the instance of Video

        The custom function can manipulate actors and camera in the scene and 
        add frames to the video with 'video.addFrame()'. 
        Once all frames are ready it has to return the video object 
        so that the video can be closed and saved.     

        :param video_function: custom function used to generate the video's frames

        see: examples/advanced/custom_videomaker.py

        """
        self.parse_kwargs(**kwargs)
        os.chdir(self.save_fld)

        # Create video
        video = Video(
            name=self.save_name, duration=self.duration, fps=self.fps
        )

        # run custom function
        video = video_function(scene=self.scene, video=video, videomaker=self)

        # Check output
        if video is None or not isinstance(video, Video):
            raise ValueError(
                "The custom video function didn't return anything "
                + "or it returned something other than the instance of Video "
                + "It must return the video object so that it can be closed properly."
            )
        if not isinstance(video, Video):
            raise ValueError(
                f"The custom video function returned invalid objects: {video} instead of video object"
            )

        # close video
        video.close()



import os
import numpy as np

from vedo import merge
from vedo.mesh import Mesh as Actor
from vedo import settings

from morphapi.morphology.morphology import Neuron

import brainrender
from brainrender.scene import Scene
from brainrender.colors import colorMap, getColor, _mapscales_cmaps
from brainrender.morphology.utils import get_neuron_actors_with_morphapi
from brainrender.Utils.data_manipulation import return_list_smart


class MorphologyScene(Scene):

    _default_axes_params = dict(
        xyGrid=True,
        yzGrid=True,
        zxGrid=True,
        xyPlaneColor="k",
        zxPlaneColor=[0.2, 0.2, 0.2],
        yzPlaneColor=[0.2, 0.2, 0.2],
        # xyGridColor = 'red',
        # xyAlpha = 1,
        # # axesLineWidth = 10,
        gridLineWidth=0,
        xyGridColor="k",
        zxGridColor="k",
        yzGridColor="k",
    )

    def __init__(self, *args, **kwargs):
        self.default_neuron_color = kwargs.pop(
            "default_neuron_color", "darksalmon"
        )
        show_axes = kwargs.pop("show_axes", True)
        axes_kwargs = kwargs.pop("axes_kwargs", 1)
        settings.DEFAULT_NEURITE_RADIUS = ("neurite_radius", 18)

        if axes_kwargs == 1:
            settings.useDepthPeeling = (
                False  # necessary to make the axes render properly
            )
            settings.useFXAA = False

        # Initialise scene class
        Scene.__init__(
            self, add_root=False, display_inset=False, *args, **kwargs
        )

        if show_axes:
            brainrender.SHOW_AXES = True
            if axes_kwargs == 1:
                self.plotter.axes = self._default_axes_params

            else:
                self.plotter.axes = axes_kwargs

    def _add_neurons_get_colors(self, neurons, color):
        """     
            Parses color argument for self.add_neurons
            
            :para, neurons: list of Neuron object or file paths...
            :param color: default None. Can be:
                - None: each neuron is colored according to the default color
                - color: rbg, hex etc. If a single color is passed all neurons will have that color
                - cmap: str with name of a colormap: neurons are colored based on their sequential order and cmap
                - dict: a dictionary specifying a color for soma, dendrites and axon actors, will be the same for all neurons
                - list: a list of length = number of neurons with either a single color for each neuron
                        or a dictionary of colors for each neuron
        """
        N = len(neurons)
        colors = dict(soma=None, axon=None, dendrites=None,)

        # If no color is passed, get random colors
        if color is None:
            cols = [self.default_neuron_color for n in np.arange(N)]
            colors = dict(
                soma=cols.copy(), axon=cols.copy(), dendrites=cols.copy(),
            )
        else:
            if isinstance(color, str):
                # Deal with a a cmap being passed
                if color in _mapscales_cmaps:
                    cols = [
                        colorMap(n, name=color, vmin=-2, vmax=N + 2)
                        for n in np.arange(N)
                    ]
                    colors = dict(
                        soma=cols.copy(),
                        axon=cols.copy(),
                        dendrites=cols.copy(),
                    )

                else:
                    # Deal with a single color being passed
                    cols = [getColor(color) for n in np.arange(N)]
                    colors = dict(
                        soma=cols.copy(),
                        axon=cols.copy(),
                        dendrites=cols.copy(),
                    )
            elif isinstance(color, dict):
                # Deal with a dictionary with color for each component
                if "soma" not in color.keys():
                    raise ValueError(
                        f"When passing a dictionary as color argument, \
                                                soma should be one fo the keys: {color}"
                    )
                dendrites_color = color.pop("dendrites", color["soma"])
                axon_color = color.pop("axon", color["soma"])

                colors = dict(
                    soma=[color["soma"] for n in np.arange(N)],
                    axon=[axon_color for n in np.arange(N)],
                    dendrites=[dendrites_color for n in np.arange(N)],
                )

            elif isinstance(color, (list, tuple)):
                # Check that the list content makes sense
                if len(color) != N:
                    raise ValueError(
                        "When passing a list of color arguments, the list length"
                        + f" ({len(color)}) should match the number of neurons ({N})."
                    )
                if len(set([type(c) for c in color])) > 1:
                    raise ValueError(
                        "When passing a list of color arguments, all list elements"
                        + " should have the same type (e.g. str or dict)"
                    )

                if isinstance(color[0], dict):
                    # Deal with a list of dictionaries
                    soma_colors, dendrites_colors, axon_colors = [], [], []

                    for col in color:
                        if "soma" not in col.keys():
                            raise ValueError(
                                f"When passing a dictionary as col argument, \
                                                        soma should be one fo the keys: {col}"
                            )
                        dendrites_colors.append(
                            col.pop("dendrites", col["soma"])
                        )
                        axon_colors.append(col.pop("axon", col["soma"]))
                        soma_colors.append(col["soma"])

                    colors = dict(
                        soma=soma_colors,
                        axon=axon_colors,
                        dendrites=dendrites_colors,
                    )

                else:
                    # Deal with a list of colors
                    colors = dict(
                        soma=color.copy(),
                        axon=color.copy(),
                        dendrites=color.copy(),
                    )
            else:
                raise ValueError(
                    f"Color argument passed is not valid. Should be a \
                                        str, dict, list or None, not {type(color)}:{color}"
                )

        # Check colors, if everything went well we should have N colors per entry
        for k, v in colors.items():
            if len(v) != N:
                raise ValueError(
                    f"Something went wrong while preparing colors. Not all \
                                entries have right length. We got: {colors}"
                )

        return colors

    def add_neurons(
        self,
        neurons,
        color=None,
        display_axon=True,
        display_dendrites=True,
        alpha=1,
        neurite_radius=None,
    ):
        """
            Adds rendered morphological data of neurons reconstructions downloaded from the
            Mouse Light project at Janelia, neuromorpho.org and other sources. 
            Accepts neurons argument as:
                - file(s) with morphological data
                - vedo mesh actor(s) of neurons reconstructions
                - dictionary or list of dictionary with actors for different neuron parts

            :param self: instance of brainrender Scene to use to render neurons
            :param neurons: str, list, dict. File(s) with neurons data or list of rendered neurons.
            :param display_axon, display_dendrites: if set to False the corresponding neurite is not rendered
            :param color: default None. Can be:
                    - None: each neuron is colored according to the default color
                    - color: rbg, hex etc. If a single color is passed all neurons will have that color
                    - cmap: str with name of a colormap: neurons are colored based on their sequential order and cmap
                    - dict: a dictionary specifying a color for soma, dendrites and axon actors, will be the same for all neurons
                    - list: a list of length = number of neurons with either a single color for each neuron
                            or a dictionary of colors for each neuron
            :param alpha: float in range 0,1. Neurons transparency
            :param neurite_radius: float > 0 , radius of tube actor representing neurites
        """

        if not isinstance(neurons, (list, tuple)):
            neurons = [neurons]

        # ------------------------------ Prepare colors ------------------------------ #
        colors = self._add_neurons_get_colors(neurons, color)

        # ---------------------------------- Render ---------------------------------- #
        _neurons_actors = []
        for neuron in neurons:
            neuron_actors = {"soma": None, "dendrites": None, "axon": None}

            # Deal with neuron as filepath
            if isinstance(neuron, str):
                if os.path.isfile(neuron):
                    if neuron.endswith(".swc"):
                        neuron_actors, _ = get_neuron_actors_with_morphapi(
                            swcfile=neuron, neurite_radius=neurite_radius
                        )
                    else:
                        raise NotImplementedError(
                            "Currently we can only parse morphological reconstructions from swc files"
                        )
                else:
                    raise ValueError(
                        f"Passed neruon {neuron} is not a valid input. Maybe the file doesn't exist?"
                    )

            # Deal with neuron as single actor
            elif isinstance(neuron, Actor):
                # A single actor was passed, maybe it's the entire neuron
                neuron_actors["soma"] = neuron  # store it as soma anyway
                pass

            # Deal with neuron as dictionary of actor
            elif isinstance(neuron, dict):
                neuron_actors["soma"] = neuron.pop("soma", None)
                neuron_actors["axon"] = neuron.pop("axon", None)

                # Get dendrites actors
                if (
                    "apical_dendrites" in neuron.keys()
                    or "basal_dendrites" in neuron.keys()
                ):
                    if "apical_dendrites" not in neuron.keys():
                        neuron_actors["dendrites"] = neuron["basal_dendrites"]
                    elif "basal_dendrites" not in neuron.keys():
                        neuron_actors["dendrites"] = neuron["apical_dendrites"]
                    else:
                        neuron_actors["dendrites"] = merge(
                            neuron["apical_dendrites"],
                            neuron["basal_dendrites"],
                        )
                else:
                    neuron_actors["dendrites"] = neuron.pop("dendrites", None)

            # Deal with neuron as instance of Neuron from morphapi
            elif isinstance(neuron, Neuron):
                neuron_actors, _ = get_neuron_actors_with_morphapi(
                    neuron=neuron
                )
            # Deal with other inputs
            else:
                raise ValueError(
                    f"Passed neuron {neuron} is not a valid input"
                )

            # Check that we don't have anything weird in neuron_actors
            for key, act in neuron_actors.items():
                if act is not None:
                    if not isinstance(act, Actor):
                        raise ValueError(
                            f"Neuron actor {key} is {act.__type__} but should be a vedo Mesh. Not: {act}"
                        )

            if not display_axon:
                neuron_actors["axon"] = None
            if not display_dendrites:
                neuron_actors["dendrites"] = None
            _neurons_actors.append(neuron_actors)

        # Color actors
        for n, neuron in enumerate(_neurons_actors):
            if neuron["axon"] is not None:
                neuron["axon"].c(colors["axon"][n])
            neuron["soma"].c(colors["soma"][n])
            if neuron["dendrites"] is not None:
                neuron["dendrites"].c(colors["dendrites"][n])

        # Add to actors storage
        self.actors.extend([list(n.values()) for n in _neurons_actors])

        return return_list_smart(_neurons_actors)



"""
    This example shows how download neurons from the 
    mouselight datase with morphapi and render 
    them with a custom colors palette
"""

import brainrender

brainrender.SHADER_STYLE = "cartoon"
brainrender.ROOT_ALPHA = 0.3
brainrender.BACKGROUND_COLOR = "blackboard"


from brainrender.scene import Scene
from brainrender.colors import makePalette

from morphapi.api.mouselight import MouseLightAPI


# ---------------------------- Downloading neurons --------------------------- #
mlapi = MouseLightAPI()

# Fetch metadata for neurons with some in the secondary motor cortex
neurons_metadata = mlapi.fetch_neurons_metadata(
    filterby="soma", filter_regions=["MOs"]
)

# Then we can download the files and save them as a .json file
neurons = mlapi.download_neurons(
    neurons_metadata[:50]
)  # 50 neurons, might take a while the first time


# ----------------------------- Rendering neurons ---------------------------- #
# Create a custom colormap between 3 colors
colors = makePalette(len(neurons), "salmon", "lightgreen")

# Create scene
scene = Scene(add_root=True, display_inset=False, title="neurons")

# Add each neuron with it's color
scene.add_neurons(neurons, alpha=0.8, neurite_radius=8, color=colors)

# Cut all actors to expose neurons
scene.cut_actors_with_plane("sagittal")
scene.render()



""" 
    This tutorial shows how to render a neuron from a .swc file
    with the MorphologyScene class
"""
import brainrender

brainrender.SHADER_STYLE = "cartoon"

from brainrender.morphology.visualise import MorphologyScene


scene = MorphologyScene(title="A neuron")

neuron = scene.add_neurons(
    "Examples/example_files/neuron2.swc",
    color=dict(soma="red", dendrites="orangered", axon=[0.4, 0.4, 0.4],),
    neurite_radius=4,
)

scene.render()



import pandas as pd

# import numpy as np
import os
from PIL import ImageColor
import math

from vedo import load, merge, write, Points, Arrows
from vedo.shapes import Tube
from vedo.analysis import recoSurface

from brainrender.scene import Scene
from brainrender.Utils.paths_manager import Paths
from brainrender.Utils.data_io import (
    load_mesh_from_file,
    listdir,
    get_subdirs,
    load_json,
)
from brainrender.colors import get_random_colors  # , getColor
from brainrender import NEURON_RESOLUTION, ROOT_ALPHA, ROOT_COLOR

# TODO add method to select pre/post synapses based on synaptic partner


"""
This class rendereds datasets of the whole C. Elegans connectome. 
All data from Daniel Witvliet and co-authors.
Check https://www.biorxiv.org/content/10.1101/2020.04.30.066209v1 
for more details.


Note: currently there is no support to let users download the
connectome data programmatically. If you need the data please
get in touch. We are working to improve this.


check example file at Examples/custom_atlases/celegans_connectome.py
for more details

"""


class Celegans(Paths):
    atlas_name = "Celegans"
    mesh_format = "obj"  #  or obj, stl etc..

    pre_synapses_color = "darkgray"
    post_synapses_color = "blackboard"
    synapses_radius = 0.2

    skeleton_radius = 0.05

    default_camera = dict(
        position=[-15.686, 65.978, 32.901],
        focal=[13.312, 20.159, -9.482],
        viewup=[-0.896, -0.412, -0.168],
        distance=68.823,
        clipping=[26.154, 122.7],
    )

    def __init__(self, data_folder=None, base_dir=None, **kwargs):
        """
            This class handles loading and parsing neuroanatomical data for the C. elegans connectome from 
            https://www.biorxiv.org/content/10.1101/2020.04.30.066209v1

            :param base_dir: path to directory to use for saving data (default value None)
            :param kwargs: can be used to pass path to individual data folders. See brainrender/Utils/paths_manager.py
            :param data_folder: str, path to a folder with data for the connectome # TODO replace with downloading data
        """
        # Initialise atlas
        Paths.__init__(self, base_dir, **kwargs)

        # Get data
        if data_folder is None:
            raise ValueError(
                "No data folder was passed, use the 'atlas_kwargs' argument of Scene to pass a data folder path"
            )
        if not os.path.isdir(data_folder):
            raise FileNotFoundError(f"The folder {data_folder} does not exist")
        self.data_folder = data_folder
        self._get_data()

    def _make_root(self, rootpath):
        """
            Creates a root mesh by merging the mesh corresponding to each neuron,
            then saves it as an obj file at rootpath
        """
        raise NotImplementedError(
            "Create root method not supported yet, sorry"
        )

        print(f"Creating root mesh for atlas {self.atlas_name}")
        temp_scene = Scene(
            atlas=Celegans,
            add_root=False,
            display_inset=False,
            atlas_kwargs=dict(data_folder=self.data_folder),
        )

        temp_scene.add_neurons(self.neurons_names)
        temp_scene.render(interactive=False)
        temp_scene.close()

        root = merge(*temp_scene.actors["neurons"]).clean().cap()
        # root = mesh2Volume(root, spacing=(0.02, 0.02, 0.02)).isosurface()

        points = Points(root.points()).smoothMLS2D(f=0.8).clean(tol=0.005)

        root = recoSurface(points, dims=100, radius=0.2)

        # Save
        write(root, rootpath)

        del temp_scene
        return root

    # ----------------------------------- Utils ---------------------------------- #
    def get_neurons_by(self, getby="pair", lookup=None):
        """
            Selects a subset of the neurons using some criteria and lookup key, 
            based on the neurons metadata

            :param getby: str, name of the metadata key to use for selecting neurons
            :param lookup: str/int.. neurons whose attribute 'getby' matches the lookup value will be selected

            :returns: list of strings with neurons names
        """
        try:  # make this work if called by a Scene class
            cs = self.atlas
        except:
            cs = self

        allowed = ["neuron", "pair", "class", "type"]

        if getby not in allowed:
            raise ValueError(
                f"Get by key should be one of {allowed} not {getby}"
            )

        filtered = list(
            cs.neurons_metadata.loc[cs.neurons_metadata[getby] == lookup][
                "neuron"
            ].values
        )
        if not filtered:
            print(f"Found 0 neurons with getby {getby} and lookup {lookup}")

        return filtered

    def get_neuron_color(self, neuron, colorby="type"):
        """
            Get a neuron's RGB color. Colors can be assigned based on different criteria
            like the neuron's type or by individual neuron etc...

            :param neuron: str, nueron name
            :param colorby: str, metadata attribute to use for coloring
            :returns: rgb values of color
        """
        try:  # make this work if called by a Scene class
            cs = self.atlas
        except:
            cs = self

        allowed = ["neuron", "individual", "ind", "pair", "class", "type"]

        if colorby not in allowed:
            raise ValueError(
                f"color by key should be one of {allowed} not {colorby}"
            )

        if colorby == "type":
            color = cs.neurons_metadata.loc[
                cs.neurons_metadata.neuron == neuron
            ]["type_color"].values[0]
            color = ImageColor.getrgb(color)
        elif (
            colorby == "individual" or colorby == "ind" or colorby == "neuron"
        ):
            color = get_random_colors()
        else:
            raise NotImplementedError

        return color

    def _get_data(self):
        """
            Loads data and metadata for the C. elegans connectome.
        """
        # Get subfolder with .obj files
        subdirs = get_subdirs(self.data_folder)
        if not subdirs:
            raise ValueError(
                "The data folder should include a subfolder which stores the neurons .obj files"
            )

        try:
            self.objs_fld = [f for f in subdirs if "objs_smoothed" in f][0]
        except:
            raise FileNotFoundError(
                "Could not find subdirectory with .obj files"
            )

        # Get filepath to each .obj
        self.neurons_files = [
            f for f in listdir(self.objs_fld) if f.lower().endswith(".obj")
        ]

        # Get synapses and skeleton files
        try:
            skeletons_file = [
                f
                for f in listdir(self.data_folder)
                if f.endswith("skeletons.json")
            ][0]
        except:
            raise FileNotFoundError(
                "Could not find file with neurons skeleton data"
            )

        try:
            synapses_file = [
                f
                for f in listdir(self.data_folder)
                if f.endswith("synapses.csv")
            ][0]
        except:
            raise FileNotFoundError("Could not find file with synapses data")

        # load data
        self.skeletons_data = load_json(skeletons_file)
        self.synapses_data = pd.read_csv(synapses_file, sep=";")

        # Get neurons metadata
        try:
            metadata_file = [
                f
                for f in listdir(self.data_folder)
                if "neuron_metadata.csv" in f
            ][0]
        except:
            raise FileNotFoundError(
                f"Could not find neurons metadata file {metadata_file}"
            )

        self.neurons_metadata = pd.read_csv(metadata_file)
        self.neurons_names = list(self.neurons_metadata.neuron.values)

    def _check_neuron_argument(self, neurons):
        """
            Checks if a list of string includes neurons name, returns only
            elements of the list that are correct names

            :param neurons: list of strings with neurons names
        """
        try:  # make this work if called by a Scene class
            cs = self.atlas
        except:
            cs = self

        if not isinstance(neurons, (list, tuple)):
            neurons = [neurons]

        good_neurons = []
        for neuron in neurons:
            if neuron not in cs.neurons_names:
                print(f"Neuron {neuron} not in dataset")
            else:
                good_neurons.append(neuron)
        return good_neurons

    def _parse_neuron_skeleton(self, neuron):
        """
            Parses a neuron's skeleton information from skeleton .json file
            to create a vtk actor that represents the neuron

            :param neuron: str, neuron name
        """
        try:  # make this work if called by a Scene class
            cs = self.atlas
        except:
            cs = self
        try:
            data = cs.skeletons_data[neuron]
        except:
            print(f"No skeleton data found for {neuron}")
            return None

        # Create an actor for each neuron's branch and then merge
        actors = []
        for branch in data["branches"]:
            coords = [data["coordinates"][str(p)] for p in branch]

            # Just like for synapses we need to adjust the coordinates to match the .obj files
            # coords are x z -y
            adjusted_coords = [(c[0], c[2], -c[1]) for c in coords]
            actors.append(
                Tube(
                    adjusted_coords,
                    r=cs.skeleton_radius,
                    res=NEURON_RESOLUTION,
                )
            )

        return merge(*actors)

    # ------------------------------- Atlas methods ------------------------------ #
    def _get_structure_mesh(self, acronym, **kwargs):
        """
            Get's the mesh for a brainregion, for this atlas it's just for
            getting/making the root mesh

        """
        if acronym != "root":
            raise ValueError(
                f"The atlas {self.atlas_name} only has one structure mesh: root. Argument {acronym} is not valid"
            )

        objpath = os.path.join(self.data_folder, "objs_smoothed", "root2.obj")
        if not os.path.isfile(objpath):
            root = self._make_root(objpath)
        else:
            root = load(objpath)

        root.c(ROOT_COLOR).alpha(ROOT_ALPHA)

        return root

    def get_neurons(self, neurons, alpha=1, as_skeleton=False, colorby="type"):
        """
            Renders neurons and adds returns to the scene. 

            :param neurons: list of names of neurons
            :param alpha: float in range 0,1 -  neurons transparency
            :param as_skeleton: bool (Default value = False), if True neurons are rendered as skeletons 
                                otherwise as a full mesh showing the whole morphology
            :param colorby: str, criteria to use to color the neurons. Accepts values like type, individual etc. 
        """
        neurons = self._check_neuron_argument(neurons)

        actors, store = [], {}
        for neuron in neurons:
            if neuron not in self.neurons_names:
                print(f"Neuron {neuron} not included in dataset")
            else:
                color = self.get_neuron_color(neuron, colorby=colorby)

                if as_skeleton:  # reconstruct skeleton from json
                    actor = self._parse_neuron_skeleton(neuron)

                else:  # load as .obj file
                    try:
                        neuron_file = [
                            f for f in self.neurons_files if neuron in f
                        ][0]
                    except:
                        print(f"Could not find .obj file for neuron {neuron}")
                        continue

                    actor = load_mesh_from_file(neuron_file)

                if actor is not None:
                    # Refine actor's look
                    actor.alpha(alpha).c(color)

            actors.append(actor)
            store[neuron] = (actor, as_skeleton)

        return actors, store

    def get_neurons_synapses(
        self,
        scene_store,
        neurons,
        alpha=1,
        pre=False,
        post=False,
        colorby="synapse_type",
        draw_patches=False,
        draw_arrows=True,
    ):
        """
            THIS METHODS GETS CALLED BY SCENE, self referes to the instance of Scene not to this class.
            Renders neurons and adds them to the scene. 

            :param neurons: list of names of neurons
            :param alpha: float in range 0,1 -  neurons transparency
            :param pre: bool, if True the presynaptic sites of each neuron are rendered
            :param post: bool, if True the postsynaptic sites on each neuron are rendered
            :param colorby: str, criteria to use to color the neurons.
                             Accepts values like synapse_type, type, individual etc. 
            :param draw_patches: bool, default True. If true dark patches are used to show the location of post synapses
            :param draw_arrows: bool, default True. If true arrows are used to show the location of post synapses
        """

        col_names = ["x", "z", "y"]
        # used to correctly position synapses on .obj files

        neurons = self._check_neuron_argument(neurons)

        spheres_data, actors = [], []
        for neuron in neurons:
            if pre:
                if colorby == "synapse_type":
                    color = self.pre_synapses_color
                else:
                    color = self.get_neuron_color(neuron, colorby=colorby)

                data = self.synapses_data.loc[self.synapses_data.pre == neuron]
                if not len(data):
                    print(f"No pre- synapses found for neuron {neuron}")
                else:
                    data = data[["x", "y", "z"]]
                    data["y"] = -data["y"]

                    spheres_data.append(
                        (
                            data,
                            dict(
                                color=color,
                                verbose=False,
                                alpha=alpha,
                                radius=self.synapses_radius,
                                res=24,
                                col_names=col_names,
                            ),
                        )
                    )

            if post:
                if colorby == "synapse_type":
                    color = self.post_synapses_color
                else:
                    color = self.get_neuron_color(neuron, colorby=colorby)

                rows = [
                    i
                    for i, row in self.synapses_data.iterrows()
                    if neuron in row.posts
                ]
                data = self.synapses_data.iloc[rows]

                if not len(data):
                    print(f"No post- synapses found for neuron {neuron}")
                else:
                    data = data[["x", "y", "z"]]
                    data["y"] = -data["y"]

                    """
                        Post synaptic locations are shown as darkening of patches
                        of a neuron's mesh and or as a 3d arrow point toward the neuron.
                    """

                    spheres_data.append(
                        (
                            data,
                            dict(
                                color="black",
                                verbose=False,
                                alpha=0,
                                radius=self.synapses_radius * 4,
                                res=24,
                                col_names=col_names,
                            ),
                        )
                    )

                    # Get mesh points for neuron the synapses belong to
                    if neuron not in scene_store.keys():
                        neuron_file = [
                            f for f in self.neurons_files if neuron in f
                        ][0]
                        neuron_act = load_mesh_from_file(neuron_file, c=color)
                    else:
                        neuron_act, as_skeleton = scene_store[neuron]

                    # Draw post synapses as dark patches
                    if draw_patches:
                        if as_skeleton:
                            print(
                                "Can't display post synapses as dark spots when neron is rendered in skeleton mode"
                            )
                        else:
                            # Get faces that are inside the synapses spheres and color them darker
                            raise NotImplementedError("This needs some fixing")

                            # neuron_points = neuron_act.cellCenters()
                            # inside_points = spheres.insidePoints(
                            #     neuron_points, returnIds=True
                            # )

                            # n_cells = neuron_act.polydata().GetNumberOfCells()
                            # scals = np.zeros((n_cells))
                            # scals[inside_points] = 1

                            # colors = [
                            #     neuron_act.c()
                            #     if s == 0
                            #     else getColor("blackboard")
                            #     for s in scals
                            # ]
                            # neuron_act.cellIndividualColors(colors)

                    # Draw post synapses as arrow
                    if draw_arrows:
                        points1 = [
                            [x, y, z]
                            for x, y, z in zip(
                                data[col_names[0]].values,
                                data[col_names[1]].values,
                                data[col_names[2]].values,
                            )
                        ]

                        points2 = [neuron_act.closestPoint(p) for p in points1]

                        #  shift point1 to make arrows longer
                        def dist(p1, p2):
                            return math.sqrt(
                                (p1[0] - p2[0]) ** 2
                                + (p1[1] - p2[1]) ** 2
                                + (p1[2] - p2[2]) ** 2
                            )

                        def get_point(p1, p2, d, u):
                            alpha = (1 / d) * u
                            x = (1 - alpha) * p1[0] + alpha * p2[0]
                            y = (1 - alpha) * p1[1] + alpha * p2[1]
                            z = (1 - alpha) * p1[2] + alpha * p2[2]
                            return [x, y, z]

                        dists = [
                            dist(p1, p2) for p1, p2 in zip(points1, points2)
                        ]
                        points0 = [
                            get_point(p1, p2, d, -0.5)
                            for p1, p2, d in zip(points1, points2, dists)
                        ]

                        arrows = Arrows(
                            points0, endPoints=points2, c=color, s=4
                        )

                        # ! aasduaisdbasiudbuia
                        actors.append(arrows)
        return spheres_data, actors



""" 
    This tutorial shows how to create regions with specific camera parameters
"""
import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

# Create a scene
scene = Scene(
    camera="top", title="camera"
)  # specify that you want a view from the top

# render
scene.render()

# Now render but with a different view
scene.render(camera="sagittal", zoom=1, title="camera")

# Now render but with specific camera parameters
bespoke_camera = dict(
    position=[801.843, -1339.564, 8120.729],
    focal=[9207.34, 2416.64, 5689.725],
    viewup=[0.36, -0.917, -0.171],
    distance=9522.144,
    clipping=[5892.778, 14113.736],
)

scene.render(camera=bespoke_camera)



import os
from pathlib import Path
import sys
import brainrender.default_variables
from brainrender.Utils.data_io import save_yaml, load_yaml
from brainrender.Utils.ruler import ruler
import warnings

import pyinspect

pyinspect.install_traceback()

__all__ = [
    "DEFAULT_ATLAS",
    "BACKGROUND_COLOR",
    "DEFAULT_HDF_KEY",
    "DEFAULT_NEURITE_RADIUS",
    "DEFAULT_STRUCTURE_ALPHA",
    "DEFAULT_STRUCTURE_COLOR",
    "DISPLAY_INSET",
    "DISPLAY_ROOT",
    "HDF_SUFFIXES",
    "INJECTION_DEFAULT_COLOR",
    "INJECTION_VOLUME_SIZE",
    "NEURON_ALPHA",
    "NEURON_RESOLUTION",
    "ROOT_ALPHA",
    "ROOT_COLOR",
    "SHADER_STYLE",
    "SHOW_AXES",
    "SOMA_RADIUS",
    "STREAMLINES_RESOLUTION",
    "TRACTO_ALPHA",
    "TRACTO_RADIUS",
    "TRACTO_RES",
    "TRACT_DEFAULT_COLOR",
    "VERBOSE",
    "WHOLE_SCREEN",
    "WINDOW_POS",
    "INTERACTIVE_MSG",
    "CAMERA",
    "DEFAULT_SCREENSHOT_NAME",
    "SCREENSHOT_TRANSPARENT_BACKGROUND",
    "AXES_STYLE",
]


# -------------------------- Set vedo shaders -------------------------- #
from vedo import settings

settings.useDepthPeeling = (
    True  # necessary for rendering of semitransparent actors
)
settings.useFXAA = True  # necessary for rendering of semitransparent actors


# ------------------------- reset default parameters file ------------------------- #
params_file = Path(os.path.expanduser("~")) / ".brainrender" / "config.yaml"
defaults = brainrender.default_variables.__dict__
comment = (
    "# Rendering options. An explanation for each parameter can be found "
    + "in the documentation or in brainrender.default_variables.py\n"
)


def reset_defaults():
    # Get all variables from defaults
    vs = {
        key: value
        for key, value in defaults.items()
        if not (key.startswith("__") or key.startswith("_"))
    }
    save_yaml(str(params_file), vs, append=False, topcomment=comment)


# ---------------------------------------------------------------------------- #
#                Create a config file from the default variables               #
# ---------------------------------------------------------------------------- #
# Get base directory
_user_dir = Path(os.path.expanduser("~"))
if not _user_dir.exists():
    raise FileExistsError(
        "Could not find user base folder (to save brainrender data). Platform: {}".format(
            sys.platform
        )
    )
_base_dir = _user_dir / ".brainrender"
_base_dir.mkdir(exist_ok=True)


# Create config path
_config_path = _base_dir / "config.yaml"
if not _config_path.exists():
    reset_defaults()


# ---------------------------------------------------------------------------- #
#                                PARSE VARIABLES                               #
# ---------------------------------------------------------------------------- #
# Rendering options. An explanation for each parameter can be found in the documentation or in brainrender.default_variables.py
params = load_yaml(str(_config_path))

# Check we have all the params
for par in __all__:
    if par in ["INTERACTIVE_MSG"]:
        continue
    if par not in params.keys():
        params[par] = defaults[par]
save_yaml(str(params_file), params, append=False, topcomment=comment)

# ------------------------- Other vedo settings ------------------------ #
if params["SCREENSHOT_TRANSPARENT_BACKGROUND"]:
    settings.screenshotTransparentBackground = True  # vedo for transparent bg
    settings.useFXAA = False  # This needs to be false for transparent bg


# Set to make it easy to import
AXES_STYLE = params["AXES_STYLE"]
DEFAULT_ATLAS = params["DEFAULT_ATLAS"]
BACKGROUND_COLOR = params["BACKGROUND_COLOR"]
DEFAULT_HDF_KEY = params["DEFAULT_HDF_KEY"]
DEFAULT_NEURITE_RADIUS = params["DEFAULT_NEURITE_RADIUS"]
DEFAULT_STRUCTURE_ALPHA = params["DEFAULT_STRUCTURE_ALPHA"]
DEFAULT_STRUCTURE_COLOR = params["DEFAULT_STRUCTURE_COLOR"]
DISPLAY_INSET = params["DISPLAY_INSET"]
DISPLAY_ROOT = params["DISPLAY_ROOT"]
HDF_SUFFIXES = params["HDF_SUFFIXES"]
INJECTION_DEFAULT_COLOR = params["INJECTION_DEFAULT_COLOR"]
INJECTION_VOLUME_SIZE = params["INJECTION_VOLUME_SIZE"]
NEURON_ALPHA = params["NEURON_ALPHA"]
NEURON_RESOLUTION = params["NEURON_RESOLUTION"]
ROOT_ALPHA = params["ROOT_ALPHA"]
ROOT_COLOR = params["ROOT_COLOR"]
SHADER_STYLE = params["SHADER_STYLE"]
SHOW_AXES = params["SHOW_AXES"]
SOMA_RADIUS = params["SOMA_RADIUS"]
STREAMLINES_RESOLUTION = params["STREAMLINES_RESOLUTION"]
TRACTO_ALPHA = params["TRACTO_ALPHA"]
TRACTO_RADIUS = params["TRACTO_RADIUS"]
TRACTO_RES = params["TRACTO_RES"]
TRACT_DEFAULT_COLOR = params["TRACT_DEFAULT_COLOR"]
VERBOSE = params["VERBOSE"]
WHOLE_SCREEN = params["WHOLE_SCREEN"]
WINDOW_POS = params["WINDOW_POS"]
CAMERA = params["CAMERA"]
DEFAULT_SCREENSHOT_NAME = params["DEFAULT_SCREENSHOT_NAME"]
SCREENSHOT_TRANSPARENT_BACKGROUND = params["SCREENSHOT_TRANSPARENT_BACKGROUND"]

INTERACTIVE_MSG = """
 ==========================================================
| Press: i     print info about selected object            |
|        m     minimise opacity of selected mesh           |
|        .,    reduce/increase opacity                     |
|        /     maximize opacity                            |
|        w/s   toggle wireframe/solid style                |
|        p/P   change point size of vertices               |
|        l     toggle edges line visibility                |
|        x     toggle mesh visibility                      |
|        X     invoke a cutter widget tool                 |
|        1-3   change mesh color                           |
|        4     use scalars as colors, if present           |
|        5     change background color                     |
|        0-9   (on keypad) change axes style               |
|        k     cycle available lighting styles             |
|        K     cycle available shading styles              |
|        o/O   add/remove light to scene and rotate it     |
|        n     show surface mesh normals                   |
|        a     toggle interaction to Actor Mode            |
|        j     toggle interaction to Joystick Mode         |
|        r     reset camera position                       |
|        C     print current camera info                   |
|        S     save a screenshot                           |
|        E     export rendering window to numpy file       |
|        q     return control to python script             |
|        Esc   close the rendering window and continue     |
|        F1    abort execution and exit python kernel      |
| Mouse: Left-click    rotate scene / pick actors          |
|        Middle-click  pan scene                           |
|        Right-click   zoom scene in or out                |
|        Cntrl-click   rotate scene perpendicularly        |
|----------------------------------------------------------|
| Check out documentation at:  https://vedo.embl.es  |
 ==========================================================
"""



""" --------------------------------------------------------------------------------- """
# ATLAS
""" --------------------------------------------------------------------------------- """
DEFAULT_ATLAS = "allen_mouse_25um"  # default atlas

""" --------------------------------------------------------------------------------- """
# WHOLE SCENE RENDERING OPTIONS
""" --------------------------------------------------------------------------------- """
DISPLAY_INSET = (
    True  # display a small version of the brain to show the orientation,
)
# useful when the overall brian is not displayed (DISPLAY_ROOT). inset is crated at render time
DISPLAY_ROOT = True  # display the overall shape of the brain
WHOLE_SCREEN = True  # If true render window is full screen
BACKGROUND_COLOR = (
    "white"  # Secify the color of the background window (see colors.py)
)
SHOW_AXES = (
    True  # If true a triad of orthogonal axes is used to show orientation
)
AXES_STYLE = 1  # style of vedo axes
WINDOW_POS = (
    10,
    10,
)  # Position of the window in pixels from bottom left of screen. Only applies when not in fullscreen
CAMERA = "three_quarters"  # Default camera settings (orientation etc. see brainrender.camera.py)

""" --------------------------------------------------------------------------------- """
# SCREENSHOTS AND EXPORT OPTIONS
""" --------------------------------------------------------------------------------- """
DEFAULT_SCREENSHOT_NAME = "screenshot"  # screenshots will have this name and the time at which they were taken
SCREENSHOT_TRANSPARENT_BACKGROUND = (
    True  # If true the screenshots are saved with a transparent background
)

""" --------------------------------------------------------------------------------- """
# BRAIN REGIONS RENDERING OPTIONS
""" --------------------------------------------------------------------------------- """
ROOT_COLOR = [0.8, 0.8, 0.8]  # color of the overall brain model's actor
ROOT_ALPHA = 0.2  # transparency of the overall brain model's actor'
DEFAULT_STRUCTURE_COLOR = [0.8, 0.8, 0.8]
DEFAULT_STRUCTURE_ALPHA = 1


""" --------------------------------------------------------------------------------- """
# TRACTOGRAPHY & INJECTION RENDERING OPTIONS
""" --------------------------------------------------------------------------------- """
INJECTION_VOLUME_SIZE = 120  # injection locations are represented as spheres whose radius is injection-volume*INJECTION_VOLUME_SIZE
TRACTO_RADIUS = 20  # radius of tubes used to represent tracts
TRACTO_ALPHA = 1  # transparency of tracts
TRACTO_RES = 12  # resolution of tubes used to represent tracts
TRACT_DEFAULT_COLOR = "r"  # default color of tractography tubes
INJECTION_DEFAULT_COLOR = "g"  # default color for experiments injection sites
STREAMLINES_RESOLUTION = 3  # resolution of actors used to render the neuron,


""" --------------------------------------------------------------------------------- """
# MOUSE LIGHT NEURONS RENDERING VARIABLES
""" --------------------------------------------------------------------------------- """
DEFAULT_NEURITE_RADIUS = (
    8  # rneurites (axon/dendrites) radius as a fraction of soma radius
)
SOMA_RADIUS = 4  # radius of the soma sphere
NEURON_RESOLUTION = 16  # resolution of actors used to render the neuron,
NEURON_ALPHA = 0.85  # transparency of the neurons actors

""" --------------------------------------------------------------------------------- """
# OTHER VARIABLES
""" --------------------------------------------------------------------------------- """
SHADER_STYLE = "plastic"  # affects the look of rendered brain regions: [metallic, plastic, shiny, glossy]
VERBOSE = True  # if True print useful messages during use

""" --------------------------------------------------------------------------------- """
# SUPPORTED_FILE FORMATS
""" --------------------------------------------------------------------------------- """
HDF_SUFFIXES = [".h5", ".hdf", ".hdf5", ".he5"]
DEFAULT_HDF_KEY = "hdf"



from brainrender.animation.video import BasicVideoMaker as VideoMaker
from brainrender.scene import Scene

scene = Scene()

# Create an instance of VideoMaker with our scene
vm = VideoMaker(scene, niters=30)  # niters = number of frames

# Make a video!
vm.make_video(
    elevation=1, roll=5
)  # specify how the scene rotates at each frame



import numpy as np
import inspect
from vedo import Mesh, Text, Sphere
import random

from brainrender.atlases.atlas import Atlas
import brainrender
from brainrender.Utils.camera import check_camera_param
from brainrender.colors import get_random_colors


def get_scene_atlas(atlas, base_dir, atlas_kwargs={}, **kwargs):
    """
        Return an instance of an Atlas class. 

    """
    if atlas is None:
        atlas = brainrender.DEFAULT_ATLAS

    if isinstance(atlas, str):
        return Atlas(atlas, base_dir=base_dir, **atlas_kwargs, **kwargs,)
    elif inspect.isclass(atlas):
        return atlas(**atlas_kwargs)
    else:
        raise ValueError(
            "The `atlas` argument should be None, a string with atlas name or a class"
        )


def get_scene_camera(camera, atlas):
    """
        Gets a working camera. 
        In order these alternatives are used:
            - user given camera
            - atlas specific camera
            - default camera
    """
    if camera is None:
        if atlas.default_camera is not None:
            return check_camera_param(atlas.default_camera)
        else:
            return brainrender.CAMERA
    else:
        return check_camera_param(camera)


def get_scene_plotter_settings(jupyter, atlas, verbose):
    """
        Gets settings for vedo Plotter

    """

    if brainrender.WHOLE_SCREEN and not jupyter:
        sz = "full"
    elif brainrender.WHOLE_SCREEN and jupyter:
        if verbose:
            print(
                "Setting window size to 'auto' as whole screen is not available in jupyter"
            )
        sz = "auto"
    else:
        sz = "auto"

    if brainrender.SHOW_AXES:
        if brainrender.AXES_STYLE == 1:
            try:
                ax_idx = atlas.space.axes_order.index("frontal")
            except AttributeError:
                # some custom atlases might not have .space
                axes = 1
            else:
                # make acustom axes dict
                atlas_shape = np.array(atlas.metadata["shape"]) * np.array(
                    atlas.metadata["resolution"]
                )

                z_ticks = [
                    (-v, str(np.abs(v).astype(np.int32)))
                    for v in np.linspace(0, atlas_shape[ax_idx], 10,)
                ]

                # make custom axes dict
                axes = dict(
                    axesLineWidth=3,
                    tipSize=0,
                    xtitle="AP (μm)",
                    ytitle="DV (μm)",
                    ztitle="LR (μm)",
                    textScale=0.8,
                    xTitleRotation=0,
                    xFlipText=True,
                    zrange=np.array([-atlas_shape[2], 0]),
                    zValuesAndLabels=z_ticks,
                )

        elif brainrender.AXES_STYLE != 7:
            raise NotImplementedError(
                "Currently only AXES_STYLE=1 is supported, sorry"
            )
            axes = brainrender.AXES_STYLE
        else:
            # we need to make our custom axes in Render
            axes = 0
    else:
        axes = 0

    return dict(
        size=sz, axes=axes, pos=brainrender.WINDOW_POS, title="brainrender"
    )


def get_cells_colors_from_metadata(color_by_metadata, coords_df, color):
    """
        Get color of each cell given some metadata entry

        :param color_by_metadata: str, column name with metadata info
        :param coords_df: dataframe with cell coordinates and metadata
    """

    if color_by_metadata not in coords_df.columns:
        raise ValueError(
            'Color_by_metadata argument should be the name of one of the columns of "coords"'
        )

    # Get a map from metadata values to colors
    vals = list(coords_df[color_by_metadata].values)
    if len(vals) == 0:
        raise ValueError(
            f"Cant color by {color_by_metadata} as no values were found"
        )
    if not isinstance(
        color, dict
    ):  # The user didn't pass a lookup, generate random
        base_cols = get_random_colors(n_colors=len(set(vals)))
        cols_lookup = {v: c for v, c in zip(set(vals), base_cols)}
    else:
        try:
            for val in list(set(vals)):
                color[val]
        except KeyError:
            raise ValueError(
                'While using "color_by_metadata" with a dictionary of colors passed'
                + ' to "color", not every metadata value was assigned a color in the dictionary'
                + " please make sure that the color dictionary is complete"
            )
        else:
            cols_lookup = color

    # Use the map to get a color for each cell
    color = [cols_lookup[v] for v in vals]

    return color


def make_actor_label(
    atlas,
    actors,
    labels,
    size=300,
    color=None,
    radius=100,
    xoffset=0,
    yoffset=0,
    zoffset=0,
):
    """
        Adds a 2D text ancored to a point on the actor's mesh
        to label what the actor is

        :param kwargs: key word arguments can be passed to determine 
                text appearance and location:
                    - size: int, text size. Default 300
                    - color: str, text color. A list of colors can be passed
                            if None the actor's color is used. Default None.
                    - xoffset, yoffset, zoffset: integers that shift the label position
                    - radius: radius of sphere used to denote label anchor. Set to 0 or None to hide. 
    """
    # Check args
    if not isinstance(actors, (tuple, list)):
        actors = [actors]
    if not isinstance(labels, (tuple, list)):
        labels = [labels]

    if atlas.atlas_name == "ABA":
        offset = [-yoffset, -zoffset, xoffset]
        default_offset = np.array([0, -200, 100])
    else:
        offset = [xoffset, yoffset, zoffset]
        default_offset = np.array([100, 0, -200])

    new_actors = []
    for n, (actor, label) in enumerate(zip(actors, labels)):
        if not isinstance(actor, Mesh):
            raise ValueError(
                f"Mesh must be an instance of Mesh, not {type(actor)}"
            )
        if not isinstance(label, str):
            raise ValueError(f"Label must be a string, not {type(label)}")

        # Get label color
        if color is None:
            color = actor.c()
        elif isinstance(color, (list, tuple)):
            color = color[n]

        # Get mesh's highest point
        points = actor.points().copy()
        point = points[np.argmin(points[:, 1]), :]
        point += np.array(offset) + default_offset

        try:
            if atlas.hemisphere_from_coords(point, as_string=True) == "left":
                point = atlas.mirror_point_across_hemispheres(point)
        except IndexError:
            pass

        # Create label
        txt = Text(label, point, s=size, c=color)
        new_actors.append(txt)

        # Mark a point on Mesh that corresponds to the label location
        if radius is not None:
            pt = actor.closestPoint(point)
            new_actors.append(Sphere(pt, r=radius, c=color))

    return new_actors


def get_n_random_points_in_region(atlas, region, N, hemisphere=None):
    """
    Gets N random points inside (or on the surface) of the mesh defining a brain region.

    :param region: str, acronym of the brain region.
    :param N: int, number of points to return.
    """
    if isinstance(region, Mesh):
        region_mesh = region
    else:
        if hemisphere is None:
            region_mesh = atlas._get_structure_mesh(region)
        else:
            region_mesh = atlas.get_region_unilateral(
                region, hemisphere=hemisphere
            )
        if region_mesh is None:
            return None

    region_bounds = region_mesh.bounds()

    X = np.random.randint(region_bounds[0], region_bounds[1], size=10000)
    Y = np.random.randint(region_bounds[2], region_bounds[3], size=10000)
    Z = np.random.randint(region_bounds[4], region_bounds[5], size=10000)
    pts = [[x, y, z] for x, y, z in zip(X, Y, Z)]

    try:
        ipts = region_mesh.insidePoints(pts).points()
    except:
        ipts = region_mesh.insidePoints(
            pts
        )  # to deal with older instances of vedo
    return random.choices(ipts, k=N)



import pytest


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", help="run slow tests")


def pytest_runtest_setup(item):
    if "slow" in item.keywords and not item.config.getvalue("runslow"):
        pytest.skip("need --runslow option to run")



from pathlib import Path
import json
import yaml
import gzip
import numpy as np
import scipy.io as sio
import pandas as pd
from brainio import brainio
from vedo import load
import brainrender
from brainrender.Utils.decorators import check_file_exists
from accepts import accepts

# ------------------------------------ OS ------------------------------------ #
@accepts((str, Path))
def listdir(fld):
    """
    List the files into a folder with the complete file path instead of the relative file path like os.listdir.

    :param fld: string, folder path

    """
    return [str(f) for f in Path(fld).glob("**/*") if f.is_file()]


@accepts((str, Path))
def get_subdirs(folderpath):
    """
        Returns the subfolders in a given folder
    """
    return [str(f) for f in Path(folderpath).glob("**/*") if f.is_dir()]


# ------------------------------ Load/Save data ------------------------------ #
@check_file_exists
@accepts((str, Path), str)
def load_cells_from_file(filepath, hdf_key="hdf"):
    csv_suffix = ".csv"
    supported_formats = brainrender.HDF_SUFFIXES + [csv_suffix]

    #  check that the filepath makes sense
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(filepath)

    # check that the file is of the supported types
    if filepath.suffix == csv_suffix:
        cells = pd.read_csv(filepath)

    elif filepath.suffix in supported_formats:
        # parse file and load cell locations
        try:
            # Try reading without hdf key
            cells = pd.read_hdf(filepath)

        except KeyError:
            # Try reading with hdf key
            if filepath.suffix in brainrender.HDF_SUFFIXES:
                if hdf_key is None:
                    hdf_key = brainrender.DEFAULT_HDF_KEY
                try:
                    cells = pd.read_hdf(filepath, key=hdf_key)
                except KeyError:
                    if hdf_key == brainrender.DEFAULT_HDF_KEY:
                        raise ValueError(
                            f"The default identifier: {brainrender.DEFAULT_HDF_KEY} "
                            f"cannot be found in the hdf file. Please supply "
                            f"a key using 'scene.add_cells_from_file(filepath, "
                            f"hdf_key='key'"
                        )
                    else:
                        raise ValueError(
                            f"The key: {hdf_key} cannot be found in the hdf "
                            f"file. Please check the correct identifer."
                        )

    elif filepath.suffix == ".pkl":
        cells = pd.read_pikle(filepath)

    else:
        raise NotImplementedError(
            f"File format: {filepath.suffix} is not currently supported. "
            f"Please use one of: {supported_formats + ['.pkl']}"
        )

    return cells, filepath.name


@check_file_exists
@accepts((str, Path))
def load_npy_from_gz(filepath):
    f = gzip.GzipFile(filepath, "r")
    return np.load(f)


@accepts((str, Path))
def save_npy_to_gz(filepath, data):
    f = gzip.GzipFile(filepath, "w")
    np.save(f, data)
    f.close()


@accepts((str, Path), dict, bool, (type(None), str, dict))
def save_yaml(filepath, content, append=False, topcomment=None):
    """
    Saves content to a yaml file

    :param filepath: path to a file (must include .yaml)
    :param content: dictionary of stuff to save

    """
    if "yaml" not in filepath:
        raise ValueError("filepath is invalid")

    if not append:
        method = "w"
    else:
        method = "w+"

    with open(filepath, method) as yaml_file:
        if topcomment is not None:
            yaml_file.write(topcomment)
        yaml.dump(content, yaml_file, default_flow_style=False, indent=4)


@check_file_exists
@accepts((str, Path))
def load_json(filepath):
    """
    Load a JSON file

    :param filepath: path to a file

    """
    with open(filepath) as f:
        data = json.load(f)
    return data


@check_file_exists
@accepts((str, Path))
def load_yaml(filepath):
    """
    Load a YAML file

    :param filepath: path to yaml file

    """
    return yaml.load(open(filepath), Loader=yaml.FullLoader)


@check_file_exists
@accepts((str, Path))
def load_volume_file(filepath):
    """
    Load a volume file (e.g., .nii) and returns the data

    :param filepath: path to file
    :param **kwargs: 

    """
    try:
        volume = brainio.load_any(filepath)
    except Exception as e:
        raise ValueError(f"Could not load volume data: {filepath}:\n {e}")
    else:
        return volume


@check_file_exists
def load_mesh_from_file(filepath, *args, **kwargs):
    """	
    Load a a mesh or volume from files like .obj, .stl, ...

    :param filepath: path to file
    :param **kwargs: 

    """
    actor = load(str(filepath))
    color = kwargs.pop("color", None)
    alpha = kwargs.pop("alpha", None)

    if color is not None:
        actor.c(color)
    if alpha is not None:
        actor.alpha(alpha)

    return actor


# ----------------------------- Data manipulation ---------------------------- #
@check_file_exists
@accepts((str, Path), (int, float))
def get_probe_points_from_sharptrack(points_filepath, scale_factor=10):
    """
        Loads the location of the of probe points as extracted by SharpTrack
        [https://github.com/cortex-lab/allenCCF].

        :param points_filepath: str, path to a .mat file with probe points
        :param scale_factor: 10, sharptrack uses a 10um reference atlas so the 
                coordinates need to be scaled to match brainrender's
    """
    probe_points = sio.loadmat(points_filepath)
    probe_points = probe_points["pointList"][0][0][0][0][0]
    probe_points_df = pd.DataFrame(
        dict(
            x=probe_points[:, 2] * scale_factor,
            y=probe_points[:, 1] * scale_factor,
            z=probe_points[:, 0] * scale_factor,
        )
    )
    return probe_points_df



from brainrender import DEFAULT_NEURITE_RADIUS, SOMA_RADIUS
from morphapi.morphology.morphology import Neuron
from vedo import merge


def get_neuron_actors_with_morphapi(
    swcfile=None,
    neuron=None,
    neurite_radius=None,
    use_cache=True,
    soma_radius=None,
):
    if swcfile is None and neuron is None:
        raise ValueError("No input passed")

    if swcfile is not None:
        neuron = Neuron(data_file=swcfile)

    if neurite_radius is None:
        neurite_radius = DEFAULT_NEURITE_RADIUS
    if soma_radius is None:
        soma_radius = SOMA_RADIUS

    actors = neuron.create_mesh(
        neurite_radius=neurite_radius,
        use_cache=use_cache,
        soma_radius=soma_radius,
    )
    if actors is None:
        raise ValueError(f"Failed to get neuron actors. {swcfile} - {neuron}")
    else:
        neurites, whole_neuron = actors

    actors = dict(
        soma=neurites["soma"],
        axon=neurites["axon"],
        dendrites=merge(
            neurites["basal_dendrites"], neurites["apical_dendrites"]
        ),
    )

    return actors, whole_neuron



import numpy as np
import pandas as pd
from rich.progress import track
from vedo import shapes

try:
    from allensdk.api.queries.mouse_connectivity_api import (
        MouseConnectivityApi,
    )

    allen_sdk_installed = True
except ModuleNotFoundError:
    allen_sdk_installed = False

import brainrender
from brainrender.ABA.aba_utils import (
    parse_streamline,
    download_streamlines,
    experiments_source_search,
    parse_tractography_colors,
)
from brainrender.Utils.data_manipulation import is_any_item_in_list


class ABA:
    """
        This class augments the functionality of
        BrainGlobeAtlas with methods specific to the Allen
        Mouse Brain atlas and necessary to populate scenes in 
        brainrender. These include stuff like fetching streamlines
        and neuronal morphology data. 
    """

    atlas_name = "ABA"

    excluded_regions = ["fiber tracts"]

    # Used for streamlines
    base_url = "https://neuroinformatics.nl/HBP/allen-connectivity-viewer/json/streamlines_NNN.json.gz"

    def __init__(self):
        # mouse connectivity API [used for tractography]
        if allen_sdk_installed:
            self.mca = MouseConnectivityApi()
        else:
            self.mca = None

    # ------------------------- Scene population methods ------------------------- #

    def get_tractography(
        self,
        tractography,
        color=None,
        color_by="manual",
        others_alpha=1,
        verbose=True,
        VIP_regions=[],
        VIP_color=None,
        others_color="white",
        include_all_inj_regions=False,
        display_injection_volume=True,
    ):
        """
        Renders tractography data and adds it to the scene. A subset of tractography data can receive special treatment using the  with VIP regions argument:
        if the injection site for the tractography data is in a VIP regions, this is colored differently.

        :param tractography: list of dictionaries with tractography data
        :param color: color of rendered tractography data

        :param color_by: str, specifies which criteria to use to color the tractography (Default value = "manual")
                        options:
                            -  manual, define color of each tract
                            - target_region, color by the injected region

        :param others_alpha: float (Default value = 1)
        :param verbose: bool (Default value = True)
        :param VIP_regions: list of brain regions with VIP treatement (Default value = [])
        :param VIP_color: str, color to use for VIP data (Default value = None)
        :param others_color: str, color for not VIP data (Default value = "white")
        :param include_all_inj_regions: bool (Default value = False)
        :param display_injection_volume: float, if True a spehere is added to display the injection coordinates and volume (Default value = True)
        """

        # check argument
        if not isinstance(tractography, list):
            if isinstance(tractography, dict):
                tractography = [tractography]
            else:
                raise ValueError(
                    "the 'tractography' variable passed must be a list of dictionaries"
                )
        else:
            if not isinstance(tractography[0], dict):
                raise ValueError(
                    "the 'tractography' variable passed must be a list of dictionaries"
                )

        if not isinstance(VIP_regions, list):
            raise ValueError("VIP_regions should be a list of acronyms")

        COLORS = parse_tractography_colors(
            tractography,
            include_all_inj_regions,
            color=color,
            color_by=color_by,
            VIP_regions=VIP_regions,
            VIP_color=VIP_color,
            others_color=others_color,
        )
        COLORS = [
            c
            if c is not None
            else self._get_from_structure(t["structure-abbrev"], "rgb_triplet")
            for c, t in zip(COLORS, tractography)
        ]

        # add actors to represent tractography data
        actors, structures_acronyms = [], []
        if brainrender.VERBOSE and verbose:
            print("Structures found to be projecting to target: ")

        # Loop over injection experiments
        for i, (t, color) in enumerate(zip(tractography, COLORS)):
            # Use allen metadata
            if include_all_inj_regions:
                inj_structures = [
                    x["abbreviation"] for x in t["injection-structures"]
                ]
            else:
                inj_structures = [
                    self.get_structure_ancestors(t["structure-abbrev"])[-1]
                ]

            if (
                brainrender.VERBOSE
                and verbose
                and not is_any_item_in_list(
                    inj_structures, structures_acronyms
                )
            ):
                print("     -- ({})".format(t["structure-abbrev"]))
                structures_acronyms.append(t["structure-abbrev"])

            # get tractography points and represent as list
            if color_by == "target_region" and not is_any_item_in_list(
                inj_structures, VIP_regions
            ):
                alpha = others_alpha
            else:
                alpha = brainrender.TRACTO_ALPHA

            if alpha == 0:
                continue  # skip transparent ones

            # represent injection site as sphere
            if display_injection_volume:
                actors.append(
                    shapes.Sphere(
                        pos=t["injection-coordinates"],
                        c=color,
                        r=brainrender.INJECTION_VOLUME_SIZE
                        * t["injection-volume"],
                        alpha=brainrender.TRACTO_ALPHA,
                    )
                )
                actors[-1].name = (
                    str(t["injection-coordinates"]) + "_injection"
                )

            points = [p["coord"] for p in t["path"]]
            actors.append(
                shapes.Tube(
                    points,
                    r=brainrender.TRACTO_RADIUS,
                    c=color,
                    alpha=alpha,
                    res=brainrender.TRACTO_RES,
                )
            )
            actors[-1].name = str(t["injection-coordinates"]) + "_tractography"

        return actors

    def get_streamlines(self, sl_file, color=None, *args, **kwargs):
        """
        Render streamline data downloaded from https://neuroinformatics.nl/HBP/allen-connectivity-viewer/streamline-downloader.html

        :param sl_file: path to JSON file with streamliens data [or list of files]
        :param color: either a single color or a list of colors to color each streamline individually
        :param *args:
        :param **kwargs:

        """
        if not isinstance(sl_file, (list, tuple)):
            sl_file = [sl_file]

        # get a list of colors of length len(sl_file)
        if color is not None:
            if isinstance(color, (list, tuple)):
                if isinstance(color[0], (float, int)):  # it's an rgb color
                    color = [color for i in sl_file]
                elif len(color) != len(sl_file):
                    raise ValueError(
                        "Wrong number of colors, should be one per streamline or 1"
                    )
            else:
                color = [color for i in sl_file]
        else:
            color = ["salmon" for i in sl_file]

        actors = []
        if isinstance(
            sl_file[0], (str, pd.DataFrame)
        ):  # we have a list of files to add
            for slf, col in track(
                zip(sl_file, color),
                total=len(sl_file),
                description="parsing streamlines",
            ):
                if isinstance(slf, str):
                    streamlines = parse_streamline(
                        color=col, filepath=slf, *args, **kwargs
                    )
                else:
                    streamlines = parse_streamline(
                        color=col, data=slf, *args, **kwargs
                    )

                actors.extend(streamlines)
        else:
            raise ValueError(
                "unrecognized argument sl_file: {}".format(sl_file)
            )

        return actors

    # ----------------------------------- Utils ---------------------------------- #
    def get_projection_tracts_to_target(self, p0=None, **kwargs):
        """
        Gets tractography data for all experiments whose projections reach the brain region or location of iterest.
        
        :param p0: list of 3 floats with AP-DV-ML coordinates of point to be used as seed (Default value = None)
        :param **kwargs: 
        """

        if self.mca is None:
            raise ModuleNotFoundError(
                'You need allen sdk to use this functino: "pip install allensdk"'
            )

        # check args
        if p0 is None:
            raise ValueError("Please pass coordinates")
        elif isinstance(p0, np.ndarray):
            p0 = list(p0)
        elif not isinstance(p0, (list, tuple)):
            raise ValueError("Invalid argument passed (p0): {}".format(p0))

        p0 = [np.int(p) for p in p0]
        tract = self.mca.experiment_spatial_search(seed_point=p0, **kwargs)

        if isinstance(tract, str):
            raise ValueError(
                "Something went wrong with query, query error message:\n{}".format(
                    tract
                )
            )
        else:
            return tract

    def download_streamlines_for_region(self, region, *args, **kwargs):
        """
            Using the Allen Mouse Connectivity data and corresponding API, this function finds expeirments whose injections
            were targeted to the region of interest and downloads the corresponding streamlines data. By default, experiements
            are selected for only WT mice and onl when the region was the primary injection target. Look at "ABA.experiments_source_search"
            to see how to change this behaviour.

            :param region: str with region to use for research
            :param *args: arguments for ABA.experiments_source_search
            :param **kwargs: arguments for ABA.experiments_source_search

        """
        if self.mca is None:
            raise ModuleNotFoundError(
                'You need allen sdk to use this functino: "pip install allensdk"'
            )

        # Get experiments whose injections were targeted to the region
        region_experiments = experiments_source_search(
            self.mca, region, *args, **kwargs
        )
        try:
            return download_streamlines(
                region_experiments.id.values,
                streamlines_folder=self.streamlines_cache,
            )
        except:
            print(f"Could not download streamlines for region {region}")
            return [], []  # <- there were no experiments in the target region

    def download_streamlines_to_region(
        self, p0, *args, mouse_line="wt", **kwargs
    ):
        """
            Using the Allen Mouse Connectivity data and corresponding API, this function finds injection experiments
            which resulted in fluorescence being found in the target point, then downloads the streamlines data.

            :param p0: list of floats with AP-DV-ML coordinates
            :param mouse_line: str with name of the mouse line to use(Default value = "wt")
            :param *args: 
            :param **kwargs: 

        """
        experiments = pd.DataFrame(self.get_projection_tracts_to_target(p0=p0))
        if mouse_line == "wt":
            experiments = experiments.loc[experiments["transgenic-line"] == ""]
        else:
            if not isinstance(mouse_line, list):
                experiments = experiments.loc[
                    experiments["transgenic-line"] == mouse_line
                ]
            else:
                raise NotImplementedError(
                    "ops, you've found a bug!. For now you can only pass one mouse line at the time, sorry."
                )
        return download_streamlines(
            experiments.id.values, streamlines_folder=self.streamlines_cache
        )



"""
    This example shows you two way to specify at atlas other than the default one. 
"""
import brainrender
from brainrender.scene import Scene

scene = Scene(atlas="allen_human_500um", title="human atlas")
scene.render()
scene.close()

# Second method: modify default atlas
brainrender.DEFAULT_ATLAS = "allen_human_500um"
scene = Scene(title="human atlas")
scene.render()



import os
import numpy as np
import zipfile
import io
import sys
from brainrender.Utils.data_io import get_subdirs, listdir
from brainrender.Utils.webqueries import request
from brainrender.Utils.decorators import check_file_exists

# ----------------------------------- Cache ---------------------------------- #


def check_gene_cached(cache_folder, gene_id, exp_id):
    """ 
        A gene is saved in a folder in cache_folder
        with gene_id-exp_id as name. If the folder doesn't
        exist the gene is not cached.

        :param cache_folder: str, path to general cache folder for all data
        :param gene_id: str name of gene
        :param exp_id: id of experiment 
    """
    cache = [
        sub
        for sub in get_subdirs(cache_folder)
        if f"{gene_id}-{exp_id}" == os.path.basename(sub)
    ]
    if not cache:
        return False
    elif len(cache) > 1:
        raise ValueError("Found too many folders")
    else:
        return cache[0]


def download_and_cache(url, cachedir):
    """
        Given a url to download a gene's ISH experiment data, 
        this function download and unzips the data

        :param url: str, utl to download data
        :param cachedir: str, path to folder where data will be downloaded
    """
    # Get data
    req = request(url)

    # Create cache dir
    if not os.path.isdir(cachedir):
        os.mkdir(cachedir)

    # Unzip to cache dir
    z = zipfile.ZipFile(io.BytesIO(req.content))
    z.extractall(cachedir)


def load_cached_gene(cache, metric, grid_size):
    """
        Loads a gene's data from cache
    """
    files = [
        f for f in listdir(cache) if metric in f and not f.endswith(".mhd")
    ]
    if not files:
        return None
    if len(files) > 1:
        raise NotImplementedError("Deal with more than one file found")
    else:
        return read_raw(files[0], grid_size)


# --------------------------------- Open .raw -------------------------------- #
@check_file_exists
def read_raw(filepath, grid_size):
    """
        reads a .raw file with gene expression data 
        downloaded from the Allen atlas and returns 
        a numpy array with the correct grid_size.
        See as reference:
            http://help.brain-map.org/display/mousebrain/API#API-Expression3DGridsz

        :param filepath: str or Path object
    """
    filepath = str(filepath)

    # Read bytes
    with open(filepath, "rb") as test:
        content = test.read()

    # Create np array and return
    data = np.frombuffer(content, dtype="float32").reshape(grid_size)

    if sys.platform == "darwin":
        data = data.T  # TODO figure out why this is necessary on Mac OS?

    return data



import numpy as np
import os
from pathlib import Path
from vedo import Volume

# TODO see if this can be added to setup.py
try:
    from mcmodels.core import VoxelModelCache, Mask
    from mcmodels.models.voxel.voxel_connectivity_array import (
        VoxelConnectivityArray,
    )
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use this functionality you need to install mcmodels "
        + "with: 'pip install git+https://github.com/AllenInstitute/mouse_connectivity_models.git'"
    )

from brainrender.Utils.paths_manager import Paths
from brainrender.scene import Scene
from brainrender.Utils.data_io import (
    connected_to_internet,
    load_npy_from_gz,
    save_npy_to_gz,
)
from brainrender.colors import get_random_colormap


class VolumetricAPI(Paths):
    """
        This class takes care of downloading, analysing and rendering data from:
        "High-resolution data-driven model of the mouse connectome ", Knox et al 2018.
        [https://www.mitpressjournals.org/doi/full/10.1162/netn_a_00066].

        These data can be used to look at spatialised projection strength with sub-region (100um) resolution.
        e.g. to look at where in region B are the projections from region A, you can use this class.

        To download the data, this class uses code from: https://github.com/AllenInstitute/mouse_connectivity_models.
    """

    voxel_size = 100

    projections = {}
    mapped_projections = {}

    hemispheres = dict(left=1, right=2, both=3)

    def __init__(
        self,
        base_dir=None,
        add_root=True,
        use_cache=True,
        scene_kwargs={},
        **kwargs,
    ):
        """
            Initialise the class instance to get a few useful paths and variables. 

            :param base_dir: str, path to base directory in which all of brainrender data are stored. 
                    Pass only if you want to use a different one from what's default.
            :param add_root: bool, if True the root mesh is added to the rendered scene
            :param use_cache: if true data are loaded from a cache to speed things up.
                    Useful to set it to false to help debugging.
            :param scene_kwargs: dict, params passed to the instance of Scene associated with this class
        """
        Paths.__init__(self, base_dir=base_dir, **kwargs)

        # Get MCM cache
        cache_path = (
            Path(self.mouse_connectivity_volumetric)
            / "voxel_model_manifest.json"
        )

        if not cache_path.exists():
            if not connected_to_internet():
                raise ValueError(
                    "The first time you use this class it will need to download some data, but it seems that you're not connected to the internet."
                )
            print(
                "Downloading volumetric data. This will take several minutes but it only needs to be done once."
            )

        self.cache = VoxelModelCache(manifest_file=str(cache_path))
        self.voxel_array = None
        self.target_coords, self.source_coords = None, None

        # Get projection cache paths
        self.data_cache = self.mouse_connectivity_volumetric_cache
        self.data_cache_projections = os.path.join(
            self.data_cache, "projections"
        )
        self.data_cache_targets = os.path.join(self.data_cache, "targets")
        self.data_cache_sources = os.path.join(self.data_cache, "sources")

        for fold in [
            self.data_cache_projections,
            self.data_cache_targets,
            self.data_cache_sources,
        ]:
            if not os.path.isdir(fold):
                os.mkdir(fold)

        # Get structures tree
        self.structure_tree = self.cache.get_structure_tree()

        # Get scene
        self.scene = Scene(add_root=add_root, **scene_kwargs)

        # Other vars
        self.use_cache = use_cache

    def __getattr__(self, attr):
        __dict__ = super(VolumetricAPI, self).__getattribute__("__dict__")
        try:
            return __dict__["scene"].__getattribute__(attr)
        except AttributeError as e:
            raise AttributeError(
                f"Could not attribute {attr} for class VolumetricAPI:\n{e}"
            )

    # ---------------------------------------------------------------------------- #
    #                                     UTILS                                    #
    # ---------------------------------------------------------------------------- #
    # ------------------------- Interaction with mcmodels ------------------------ #

    def _get_structure_id(self, struct):
        " Get the ID of a structure (or list of structures) given it's acronym"
        if not isinstance(struct, (list, tuple)):
            struct = [struct]
        return [
            self.structure_tree.get_structures_by_acronym([s])[0]["id"]
            for s in struct
        ]

    def _load_voxel_data(self):
        "Load the VoxelData array from Knox et al 2018"
        if self.voxel_array is None:
            # Get VoxelArray
            weights_file = os.path.join(
                self.mouse_connectivity_volumetric, "voxel_model", "weights"
            )
            nodes_file = os.path.join(
                self.mouse_connectivity_volumetric, "voxel_model", "nodes"
            )

            # Try to load from numpy
            if os.path.isfile(weights_file + ".npy.gz"):
                weights = load_npy_from_gz(weights_file + ".npy.gz")
                nodes = load_npy_from_gz(nodes_file + ".npy.gz")

                # Create array
                self.voxel_array = VoxelConnectivityArray(weights, nodes)

                # Get target and source masks
                self.source_mask = self.cache.get_source_mask()
                self.target_mask = self.cache.get_target_mask()
            else:
                print("Loading voxel data, might take a few minutes.")
                # load from standard cache
                (
                    self.voxel_array,
                    self.source_mask,
                    self.target_mask,
                ) = self.cache.get_voxel_connectivity_array()

                # save to npy
                save_npy_to_gz(
                    weights_file + ".npy.gz", self.voxel_array.weights
                )
                save_npy_to_gz(nodes_file + ".npy.gz", self.voxel_array.nodes)

    def _get_coordinates_from_voxel_id(self, p0, as_source=True):
        """
            Takes the index of a voxel and returns the 3D coordinates in reference space. 
            The index number should be extracted with either a source_mask or a target_mask.
            If target_mask wa used set as_source as False.

            :param p0: int
        """
        if self.voxel_array is None:
            self._load_voxel_data()

        if as_source:
            return self.source_mask.coordinates[p0] * self.voxel_size
        else:
            return self.target_mask.coordinates[p0] * self.voxel_size

    def _get_mask_coords(self, as_source):
        if as_source:
            if self.source_coords is None:
                coordinates = self.source_mask.coordinates * self.voxel_size
                self.source_coords = coordinates
            else:
                coordinates = self.source_coords
        else:
            if self.target_coords is None:
                coordinates = self.target_mask.coordinates * self.voxel_size
                self.target_coords = coordinates
            else:
                coordinates = self.target_coords
        return coordinates

    def _get_voxel_id_from_coordinates(self, p0, as_source=True):
        if self.voxel_array is None:
            self._load_voxel_data()

        # Get the brain region from the coordinates
        coordinates = self._get_mask_coords(as_source)

        # Get the position of p0 in the coordinates volumetric array
        p0 = np.int64([round(p, -2) for p in p0])

        try:
            x_idx = (np.abs(coordinates[:, 0] - p0[0])).argmin()
            y_idx = (np.abs(coordinates[:, 1] - p0[1])).argmin()
            z_idx = (np.abs(coordinates[:, 2] - p0[2])).argmin()
            p0_idx = [x_idx, y_idx, z_idx]
        except:
            raise ValueError(
                f"Could not find the voxe corresponding to the point given: {p0}"
            )
        return p0_idx[0]

    # ----------------------------------- Cache ---------------------------------- #
    def _get_cache_filename(self, tgt, what):
        """Data are cached according to a naming convention, this function gets the name for an object
        according to the convention"""
        if what == "projection":
            fld = self.data_cache_projections
        elif what == "source":
            fld = self.data_cache_sources
        elif what == "target":
            fld = self.data_cache_targets
        else:
            raise ValueError(
                "Error while getting cached data file name.\n"
                + f"What was {what} but should be projection/source/target/actor."
            )

        name = "".join([str(i) for i in tgt])
        path = os.path.join(fld, name + ".npy.gz")
        return name, path, os.path.isfile(path)

    def _get_from_cache(self, tgt, what):
        """ tries to load objects from cached data, if they exist"""
        if not self.use_cache:
            return None

        name, cache_path, cache_exists = self._get_cache_filename(tgt, what)
        if not cache_exists:
            return None
        else:
            return load_npy_from_gz(cache_path)

    def save_to_cache(self, tgt, what, obj):
        """ Saves data to cache to avoid loading thema again in the future"""
        name, cache_path, _ = self._get_cache_filename(tgt, what)
        save_npy_to_gz(cache_path, obj)

    # ---------------------------------------------------------------------------- #
    #                                 PREPROCESSING                                #
    # ---------------------------------------------------------------------------- #

    # ------------------------- Sources and targets masks ------------------------ #

    def get_source(self, source, hemisphere="both"):
        """
            Loads the mask for a source structure

            :param source: str or list of str with acronym of source regions
            :param hemisphere: str, ['both', 'left', 'right']. Which hemisphere to consider.
        """
        if not isinstance(source, (list, tuple)):
            source = [source]

        self.source = self._get_from_cache(source, "source")
        if self.source is None:
            self._load_voxel_data()
            source_ids = self._get_structure_id(source)

            self.source = self.source_mask.get_structure_indices(
                structure_ids=source_ids,
                hemisphere_id=self.hemispheres[hemisphere],
            )
            self.save_to_cache(source, "source", self.source)
        return self.source

    def get_target_mask(self, target, hemisphere):
        """returns a 'key' array and a mask object
            used to transform projection data from linear arrays to 3D volumes.
        """
        target_ids = self._get_structure_id(target)
        self.tgt_mask = Mask.from_cache(
            self.cache,
            structure_ids=target_ids,
            hemisphere_id=self.hemispheres[hemisphere],
        )

    def get_target(self, target, hemisphere="both"):
        """
            Loads the mask for a target structure.  

            :param target: str or list of str with acronym of target regions
            :param hemisphere: str, ['both', 'left', 'right']. Which hemisphere to consider.
        """
        if not isinstance(target, (list, tuple)):
            target = [target]

        if hemisphere != "both":
            cache_name = target + [hemisphere]
        else:
            cache_name = target

        self.target = self._get_from_cache(cache_name, "target")
        if self.target is None:
            self._load_voxel_data()
            target_ids = self._get_structure_id(target)

            self.target = self.target_mask.get_structure_indices(
                structure_ids=target_ids,
                hemisphere_id=self.hemispheres[hemisphere],
            )
            self.save_to_cache(cache_name, "target", self.target)

        return self.target

    # -------------------------------- Projections ------------------------------- #

    def get_projection(
        self,
        source,
        target,
        name,
        hemisphere="both",
        projection_mode="mean",
        mode="target",
    ):
        """
                Gets the spatialised projection intensity from a source to a target. 

                :param source: str or list of str with acronym of source regions
                :param target: str or list of str with acronym of target regions
                :param name: str, name of the projection
                :param projection_mode: str, if 'mean' the data from different experiments are averaged, 
                                    if 'max' the highest value is taken.
                :param mode: str. If 'target' the spatialised projection strength in the target structures is returned, usefule
                        to see where source projects to in target. Otherwise if 'source' the spatialised projection strength in
                        the source structure is return. Useful to see which part of source projects to target.

                :return: 1D numpy array with mean projection from source to target voxels
        """
        if mode == "target":
            self.get_target_mask(target, hemisphere)
        elif mode == "source":
            self.get_target_mask(source, "right")
        else:
            raise ValueError(
                f"Invalide mode: {mode}. Should be either source or target."
            )

        cache_name = (
            sorted(source)
            + ["_"]
            + sorted(target)
            + [f"_{projection_mode}_{mode}"]
        )
        if hemisphere != "both":
            cache_name += [hemisphere]

        proj = self._get_from_cache(cache_name, "projection")
        if proj is None:
            source_idx = self.get_source(source, hemisphere)
            target_idx = self.get_target(target, hemisphere)

            self._load_voxel_data()
            projection = self.voxel_array[source_idx, target_idx]

            if mode == "target":
                axis = 0
            elif mode == "source":
                axis = 1
            else:
                raise ValueError(
                    f"Invalide mode: {mode}. Should be either source or target."
                )

            if projection_mode == "mean":
                proj = np.mean(projection, axis=axis)
            elif projection_mode == "max":
                proj = np.max(projection, axis=axis)
            else:
                raise ValueError(
                    f"Projection mode {projection_mode} not recognized.\n"
                    + 'Should be one of: ["mean", "max"].'
                )

            # Save to cache
            self.save_to_cache(cache_name, "projection", proj)
        self.projections[name] = proj
        return proj

    def get_mapped_projection(self, source, target, name, **kwargs):
        """
            Gets the spatialised projection intensity from a source to a target, but as 
            a mapped volume instead of a linear array. 

            :param source: str or list of str with acronym of source regions
            :param target: str or list of str with acronym of target regions
            :param name: str, name of the projection

            :return: 3D numpy array with projectino intensity
        """
        projection = self.get_projection(source, target, name, **kwargs)
        mapped_projection = self.tgt_mask.map_masked_to_annotation(projection)
        self.mapped_projections[name] = mapped_projection
        return mapped_projection

    def get_mapped_projection_to_point(
        self, p0, restrict_to=None, restrict_to_hemisphere="both"
    ):
        """
            Gets projection intensity from all voxels to the voxel corresponding to a point of interest
        """
        cache_name = f"proj_to_{p0[0]}_{p0[1]}_{p0[1]}"
        if restrict_to is not None:
            cache_name += f"_{restrict_to}"

        proj = self._get_from_cache(cache_name, "projection")

        if proj is None:
            p0idx = self._get_voxel_id_from_coordinates(p0, as_source=False)

            if restrict_to is not None:
                source_idx = self.get_source(
                    restrict_to, restrict_to_hemisphere
                )
                proj = self.voxel_array[source_idx, p0idx]

                self.get_target_mask(restrict_to, restrict_to_hemisphere)
                mapped_projection = self.tgt_mask.map_masked_to_annotation(
                    proj
                )
            else:
                proj = self.voxel_array[:, p0idx]
                mapped_projection = self.source_mask.map_masked_to_annotation(
                    proj
                )
            self.save_to_cache(cache_name, "projection", mapped_projection)

            return mapped_projection
        else:
            return proj

    def get_mapped_projection_from_point(
        self, p0, restrict_to=None, restrict_to_hemisphere="both"
    ):
        """
            Gets projection intensity from all voxels to the voxel corresponding to a point of interest
        """
        if self.get_hemispere_from_point(p0) == "left":
            raise ValueError(
                f"The point passed [{p0}] is in the left hemisphere,"
                + ' but "projection from point" only works from the right hemisphere.'
            )

        cache_name = f"proj_from_{p0[0]}_{p0[1]}_{p0[1]}"
        if restrict_to is not None:
            cache_name += f"_{restrict_to}"

        proj = self._get_from_cache(cache_name, "projection")

        if proj is None:
            p0idx = self._get_voxel_id_from_coordinates(p0, as_source=True)

            if restrict_to is not None:
                target_idx = self.get_target(
                    restrict_to, restrict_to_hemisphere
                )
                proj = self.voxel_array[p0idx, target_idx]

                self.get_target_mask(restrict_to, restrict_to_hemisphere)
                mapped_projection = self.tgt_mask.map_masked_to_annotation(
                    proj
                )
            else:
                proj = self.voxel_array[p0idx, :]
                mapped_projection = self.target_mask.map_masked_to_annotation(
                    proj
                )
            self.save_to_cache(cache_name, "projection", mapped_projection)

            return mapped_projection
        else:
            return proj

    # ---------------------------------------------------------------------------- #
    #                                   RENDERING                                  #
    # ---------------------------------------------------------------------------- #
    def add_mapped_projection(
        self,
        source,
        target,
        actor_kwargs={},
        render_source_region=False,
        render_target_region=False,
        regions_kwargs={},
        **kwargs,
    ):
        """
            Gets the spatialised projection intensity from a source to a target
            and renders it as a vedo lego visualisation.

            :param source: str or list of str with acronym of source regions
            :param target: str or list of str with acronym of target regions
            :param render_source_region: bool, if true a wireframe mesh of source regions is rendered
            :param render_target_region: bool, if true a wireframe mesh of target regions is rendered
            :param regions_kwargs: pass options to specify how brain regions should look like
            :param kwargs: kwargs can be used to control how the rendered object looks like. 
                    Look at the arguments of 'add_volume' to see what arguments are available. 
        """
        # Get projection data
        if not isinstance(source, list):
            source = [source]
        if not isinstance(target, list):
            target = [target]
        name = "".join(source) + "_".join(target)
        mapped_projection = self.get_mapped_projection(
            source, target, name, **kwargs
        )
        lego_actor = self.add_volume(mapped_projection, **actor_kwargs)

        # Render relevant regions meshes
        if render_source_region or render_target_region:
            wireframe = regions_kwargs.pop("wireframe", True)
            use_original_color = regions_kwargs.pop("use_original_color", True)

            if render_source_region:
                self.scene.add_brain_regions(
                    source,
                    use_original_color=use_original_color,
                    wireframe=wireframe,
                    **regions_kwargs,
                )
            if render_target_region:
                self.scene.add_brain_regions(
                    target,
                    use_original_color=use_original_color,
                    wireframe=wireframe,
                    **regions_kwargs,
                )
        return lego_actor

    def add_mapped_projection_to_point(
        self,
        p0,
        show_point=True,
        show_point_region=False,
        show_crosshair=True,
        crosshair_kwargs={},
        point_region_kwargs={},
        point_kwargs={},
        from_point=False,
        **kwargs,
    ):
        if not isinstance(p0, (list, tuple, np.ndarray)):
            raise ValueError(
                "point passed should be a list or a 1d array, not: {p0}"
            )

        restrict_to = kwargs.pop("restrict_to", None)
        restrict_to_hemisphere = kwargs.pop("restrict_to_hemisphere", "both")
        if not from_point:
            projection = self.get_mapped_projection_to_point(
                p0,
                restrict_to=restrict_to,
                restrict_to_hemisphere=restrict_to_hemisphere,
            )
        else:
            projection = self.get_mapped_projection_from_point(
                p0,
                restrict_to=restrict_to,
                restrict_to_hemisphere=restrict_to_hemisphere,
            )

        lego_actor = self.add_volume(projection, **kwargs)

        if show_point:
            color = point_kwargs.pop("color", "salmon")
            radius = point_kwargs.pop("radius", 50)
            alpha = point_kwargs.pop("alpha", 1)
            if not show_crosshair:
                self.scene.add_sphere_at_point(
                    p0, color=color, radius=radius, alpha=alpha, **point_kwargs
                )
            else:
                ml = crosshair_kwargs.pop("ml", True)
                dv = crosshair_kwargs.pop("dv", True)
                ap = crosshair_kwargs.pop("ap", True)
                self.scene.add_crosshair_at_point(
                    p0,
                    ml=ml,
                    dv=dv,
                    ap=ap,
                    line_kwargs=crosshair_kwargs,
                    point_kwargs={
                        "color": color,
                        "radius": radius,
                        "alpha": alpha,
                    },
                )

        if show_point_region:
            use_original_color = point_region_kwargs.pop(
                "use_original_color", False
            )
            alpha = point_region_kwargs.pop("alpha", 0.3)
            region = self.scene.atlas.structure_from_coords(p0)
            self.scene.add_brain_regions(
                [region],
                use_original_color=use_original_color,
                alpha=alpha,
                **point_region_kwargs,
            )

        return lego_actor

    def add_mapped_projection_from_point(self, *args, **kwargs):
        return self.add_mapped_projection_to_point(
            *args, **kwargs, from_point=True
        )

    def add_volume(
        self, volume, cmap="afmhot_r", alpha=1, add_colorbar=True, **kwargs
    ):
        """
            Renders intensitdata from a 3D numpy array as a lego volumetric actor. 

            :param volume: np 3D array with number of dimensions = those of the 100um reference space. 
            :param cmap: str with name of colormap to use
            :param alpha: float, transparency
          
            :param add_colorbar: if True a colorbar is added to show the values of the colormap
        """
        # Parse kwargs
        line_width = kwargs.pop("line_width", 1)
        if cmap == "random" or not cmap or cmap is None:
            cmap = get_random_colormap()

        # Get vmin and vmax threshold for visualisation
        vmin = kwargs.pop("vmin", 0.000001)
        vmax = kwargs.pop("vmax", np.nanmax(volume))

        # Check values
        if np.max(volume) > vmax:
            print(
                "While rendering mapped projection some of the values are above the vmax threshold."
                + "They will not be displayed."
                + f" vmax was {vmax} but found value {round(np.max(volume), 5)}."
            )

        if vmin > vmax:
            raise ValueError(
                f"The vmin threhsold [{vmin}] cannot be larger than the vmax threshold [{vmax}"
            )
        if vmin < 0:
            vmin = 0

        # Get 'lego' actor
        vol = Volume(volume)
        lego = vol.legosurface(vmin=vmin, vmax=vmax, cmap=cmap)

        # Scale and color actor
        lego.alpha(alpha).lw(line_width).scale(self.voxel_size)
        lego.cmap = cmap

        # Add colorbar
        if add_colorbar:
            lego.addScalarBar(
                vmin=vmin,
                vmax=vmax,
                horizontal=1,
                c="k",
                pos=(0.05, 0.05),
                titleFontSize=40,
            )

        # Add to scene
        actor = self.scene.add_actor(lego)
        return actor



import pytest
import pandas as pd
from vedo import Text, settings

import brainrender
from brainrender.scene import Scene, DualScene, MultiScene
from brainrender.Utils.camera import coronal_camera
from brainrender.colors import makePalette
from brainrender.Utils.scene_utils import get_n_random_points_in_region
from brainrender.Utils.data_manipulation import get_cells_in_region


@pytest.fixture
def scene():
    return Scene()


def test_default():
    brainrender.DISPLAY_INSET
    brainrender.DISPLAY_ROOT
    brainrender.WHOLE_SCREEN
    brainrender.BACKGROUND_COLOR
    brainrender.SHOW_AXES
    brainrender.WINDOW_POS
    brainrender.CAMERA
    brainrender.DEFAULT_SCREENSHOT_NAME
    brainrender.SCREENSHOT_TRANSPARENT_BACKGROUND
    brainrender.ROOT_COLOR
    brainrender.ROOT_ALPHA
    brainrender.DEFAULT_STRUCTURE_COLOR
    brainrender.DEFAULT_STRUCTURE_ALPHA
    brainrender.INJECTION_VOLUME_SIZE
    brainrender.TRACTO_RADIUS
    brainrender.TRACTO_ALPHA
    brainrender.TRACTO_RES
    brainrender.TRACT_DEFAULT_COLOR
    brainrender.INJECTION_DEFAULT_COLOR
    brainrender.STREAMLINES_RESOLUTION
    brainrender.INJECTION_VOLUME_SIZE
    brainrender.TRACTO_RADIUS
    brainrender.TRACTO_ALPHA
    brainrender.TRACTO_RES
    brainrender.TRACT_DEFAULT_COLOR
    brainrender.INJECTION_DEFAULT_COLOR
    brainrender.STREAMLINES_RESOLUTION
    brainrender.SHADER_STYLE
    brainrender.VERBOSE
    brainrender.HDF_SUFFIXES
    brainrender.DEFAULT_HDF_KEY
    brainrender.AXES_STYLE

    brainrender.reset_defaults()


def test_scene_creation():
    s = Scene()
    print(s)
    s


def test_scene_addition():
    scene = Scene()

    scene + "Examples/example_files/root.obj"

    scene + scene.root

    scene += scene.root

    # test report

    scene.list_actors()


def test_scene_creation_ignore_inset():
    Scene(display_inset=False)


def test_scene_creation_ignore_root():
    Scene(add_root=False)


def test_scene_creation_ignore_root_and_inset():
    s = Scene(add_root=False, display_inset=False)
    s.render(interactive=False)
    s.close()


def test_scene_creation_default_keys():
    Scene(use_default_key_bindings=True)


def test_scene_creation_camera():
    Scene(camera="sagittal")
    Scene(camera=coronal_camera)


def test_scene_creation_title():
    Scene(title="My title")


def test_scene_creation_notebook():
    settings.notebookBackend = "k3d"
    Scene()
    settings.notebookBackend = None


def test_scene_creation_settings():
    brainrender.SHOW_AXES = True
    brainrender.WHOLE_SCREEN = False
    Scene()
    brainrender.SHOW_AXES = False


@pytest.mark.slow
def test_scene_creation_brainglobe():
    scene = Scene(atlas="allen_mouse_25um")

    try:
        scene.add_brain_regions("TH")
    except:
        raise ValueError

    try:
        scene.add_streamlines
    except:
        raise ValueError

    scene = Scene(atlas="allen_human_500um")
    try:
        scene.add_brain_regions("TH")
    except:
        raise ValueError


def test_root(scene):
    scene.add_root()


def test_regions():
    scene = Scene(camera=coronal_camera)
    regions = ["MOs", "VISp", "ZI"]
    scene.add_brain_regions(regions, colors="green")
    ca1 = scene.add_brain_regions("CA1", add_labels=True)
    ca1.alpha(0.2)
    scene.close()


def test_cut_with_plane(scene):
    # Add some actors
    scene.add_brain_regions(["MOp"], alpha=0.5)

    # Specify position, size and orientation of the plane
    pos = scene.atlas._root_midpoint
    norm = [0, 1, 1]
    plane = scene.atlas.get_plane_at_point(pos, norm, color="lightblue")

    # Cut
    scene.cut_actors_with_plane(
        plane, close_actors=False, showplane=True, actors=scene.root,
    )
    scene.cut_actors_with_plane(
        plane, close_actors=True, showplane=False,
    )

    scene.cut_actors_with_plane(["sagittal", "frontal", "horizontal"])


def test_add_plane(scene):
    scene.add_plane(plane="sagittal")


def test_camera():
    # Create a scene
    scene = Scene(camera="top")  # specify that you want a view from the top

    # render
    scene.render(interactive=False,)
    scene.close()

    # Now render but with a different view
    scene = Scene()
    scene.render(interactive=False, camera="sagittal", zoom=1)
    scene.close()

    # Now render but with specific camera parameters
    bespoke_camera = dict(
        position=[801.843, -1339.564, 8120.729],
        focal=[9207.34, 2416.64, 5689.725],
        viewup=[0.36, -0.917, -0.171],
        distance=9522.144,
        clipping=[5892.778, 14113.736],
    )

    scene = Scene()
    scene.render(interactive=False, camera=bespoke_camera, zoom=1)
    scene.close()


def test_text_3d(scene):
    # Text to add
    s = "BRAINRENDER"

    # Specify a color for each letter
    colors = makePalette(len(s), "salmon", "powderblue")

    x = 0  # use to specify the position of each letter
    # Add letters one at the time to color them individually
    for n, letter in enumerate("BRAINRENDER"):
        if "I" == letter or "N" == letter and n < 5:  # make the spacing right
            x += 0.6
        else:
            x += 1

        # Add letter and silhouette to the scne
        act = Text(
            letter, depth=0.5, c=colors[n], pos=(x, 0, 0), justify="centered"
        )
        sil = act.silhouette().lw(3).color("k")
        scene.add_actor(act, sil)

    scene.render(interactive=False)
    scene.close()


def text_actor_labels(scene):
    # add_brain_regions can be used to add labels directly
    scene.add_brain_regions("VAL", add_labels=True)

    # you can also use scene.add_actor_label
    mos = scene.add_brain_regions("MOs")

    # Add another label, this time make it gray and shift it slightly
    scene.add_actor_label(
        mos, "MOs", size=400, color="blackboard", xoffset=250
    )


def test_crosshair(scene):
    scene.add_brain_regions(["TH"], use_original_color=False, alpha=0.4)

    # Add a point in the right hemisphere
    point = scene.atlas.get_region_CenterOfMass("TH")
    scene.add_crosshair_at_point(
        point,
        point_kwargs={
            "color": "salmon"
        },  # specify how the point at the center of the crosshair looks like
    )

    # Add a point in the left hemisphere
    point = scene.atlas.get_region_CenterOfMass("TH", hemisphere="left")
    scene.add_crosshair_at_point(
        point,
        point_kwargs={
            "color": "darkseagreen"
        },  # specify how the point at the center of the crosshair looks like
    )


def test_cells_from_file(scene):
    scene.add_cells_from_file("tests/data/cells.csv")
    scene.add_cells_from_file("tests/data/cells.h5")


def test_labelled_cells(scene):
    # Gerate the coordinates of N cells across 3 regions
    _regions = ["MOs", "VISp", "ZI"]
    N = 100  # getting 1k cells per region, but brainrender can deal with >1M cells easily.

    # Render regions
    scene.add_brain_regions(_regions, alpha=0.2)

    # Get fake cell coordinates
    cells, regions = [], []  # to store x,y,z coordinates
    for region in _regions:
        region_cells = get_n_random_points_in_region(
            scene.atlas, region=region, N=N
        )
        if len(region_cells) != N:
            raise ValueError
        cells.extend(region_cells)
        regions.extend([region for i in region_cells])
    x, y, z = (
        [c[0] for c in cells],
        [c[1] for c in cells],
        [c[2] for c in cells],
    )
    cells = pd.DataFrame(dict(x=x, y=y, z=z, region=regions))
    # Add cells
    scene.add_cells(cells, color="darkseagreen", res=12, radius=25)
    get_cells_in_region(scene.atlas, cells, "MOs")
    scene.add_cells(cells, color_by_metadata="x")
    scene.add_cells(
        cells,
        color_by_metadata="x",
        color={x: "salmon" for x in cells.x.values},
    )
    scene.add_cells(cells, color_by_region=True, radius=50, res=25)


def test_get_random_points(scene):
    get_n_random_points_in_region(scene.atlas, "MOs", 100)
    get_n_random_points_in_region(scene.atlas, "MOp", 100, hemisphere="right")
    ca1 = scene.add_brain_regions("CA1")
    get_n_random_points_in_region(scene.atlas, ca1, 100)


def test_add_from_file(scene):
    scene.add_from_file("Examples/example_files/skull.stl")


def test_add_sphere(scene):
    scene.add_sphere_at_point()


def test_add_optic_cannula(scene):
    scene.add_optic_cannula(target_region="CA1")

    p0 = scene.atlas.get_region_CenterOfMass("MOs")
    scene.add_optic_cannula(pos=p0)


def test_sharptrack(scene):
    scene.add_probe_from_sharptrack(
        "Examples/example_files/sharptrack_probe_points.mat"
    )


@pytest.mark.slow
def test_export_for_web(scene):
    scene.export_for_web()


def test_multi_scenes():
    df = DualScene()
    df.render(_interactive=False)
    df.close()

    ms = MultiScene(3)
    ms.render(_interactive=False)
    ms.close()


def test_screenshot(scene):
    scene.take_screenshot()
    scene.render(interactive=False)
    scene.take_screenshot()
    scene.close()


def test_add_actor(scene):
    act = scene.add_brain_regions("MOs")
    scene.add_actor(act)


def test_add_ruler_from_surface(scene):
    p0 = scene.atlas.get_region_CenterOfMass("TH")
    scene.add_ruler_from_surface(p0)



""" 
    This tutorial shows how to automatically take screenshots of your rendered region
"""
import time

import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

screenshot_params = dict(folder="./screenshots", name="tutorial",)


# Create a scene
scene = Scene(
    screenshot_kwargs=screenshot_params
)  # use screenshot_params to specify where the screenshots will be saved

scene.add_brain_regions(["TH"])

# render
scene.render(
    camera="top", interactive=False
)  # if interactive is false the program won't stop when the scene is rendered
# which means that the next line will be executed
scene.take_screenshot()
time.sleep(1)

# Take another screenshot from a different angle
scene.render(camera="coronal", interactive=False, zoom=0.5)
scene.take_screenshot()  # screenshots are saved with a timestamp in the name so you won't be overwriting the previous one.


# Render interactively. You can then press 's' to take a screenshot
scene.render()



"""
    This tutorial shows how to render a cross hair at a specific location in the scene
"""

from brainrender.scene import Scene

scene = Scene(title="crosshair")
scene.add_brain_regions("TH", use_original_color=False, alpha=0.4)

# Add a point in the right hemisphere
point = scene.atlas.get_region_CenterOfMass("TH")
scene.add_crosshair_at_point(
    point,
    point_kwargs={
        "color": "salmon"
    },  # specify how the point at the center of the crosshair looks like
)

# Add a point in the left hemisphere
point = scene.atlas.get_region_CenterOfMass("TH", hemisphere="left")
scene.add_crosshair_at_point(
    point,
    point_kwargs={
        "color": "darkseagreen"
    },  # specify how the point at the center of the crosshair looks like
)

scene.render()



import numpy as np
from rich.progress import track
import time
from random import choices

from brainrender.scene import Scene
from morphapi.api.mouselight import MouseLightAPI
from brainrender.Utils.camera import (
    buildcam,
    sagittal_camera,
    top_camera,
)
from brainrender.colors import colorMap

import pytest


@pytest.mark.slow
def test_animated_scene():
    # --------------------------------- Variables -------------------------------- #
    minalpha = 0.01  # transparency of background neurons
    darkcolor = "lightgray"  # background neurons color

    N_FRAMES = 50
    N_neurons = 4  # number of neurons to show in total, if -1 all neurons are shown but it might take a while to render them at first
    N_neurons_in_frame = (
        2  # number of neurons to be highlighted in a given frame
    )
    N_frames_for_change = 15  # every N frames which neurons are shown changes

    # Variables to specify camera position at each frame
    zoom = np.linspace(1, 1.5, N_FRAMES)
    frac = np.zeros_like(
        zoom
    )  # for camera transition, interpolation value between cameras
    frac[:10] = np.linspace(0, 1, 10)
    frac[10:] = np.linspace(1, 0, len(frac[10:]))

    # -------------------------------- Fetch data -------------------------------- #

    # Then we can download the files and save them as a .json file
    ml_api = MouseLightAPI()
    # Fetch metadata for neurons with some in the secondary motor cortex
    neurons_metadata = ml_api.fetch_neurons_metadata(
        filterby="soma", filter_regions=["MOs"]
    )

    neurons_files = ml_api.download_neurons(neurons_metadata[:N_neurons])

    # ------------------------------- Create scene ------------------------------- #
    scene = Scene(display_inset=False, use_default_key_bindings=True)

    neurons_actors = scene.add_neurons(
        neurons_files, neurite_radius=12, alpha=0
    )

    # Create new cameras
    cam1 = buildcam(sagittal_camera)

    cam2 = buildcam(
        dict(
            position=[-16624.081, -33431.408, 33527.412],
            focal=[6587.835, 3849.085, 5688.164],
            viewup=[0.634, -0.676, -0.376],
            distance=51996.653,
            clipping=[34765.671, 73812.327],
        )
    )

    cam3 = buildcam(
        dict(
            position=[1862.135, -4020.792, -36292.348],
            focal=[6587.835, 3849.085, 5688.164],
            viewup=[0.185, -0.97, 0.161],
            distance=42972.44,
            clipping=[29629.503, 59872.10],
        )
    )

    # ------------------------------- Create frames ------------------------------ #
    # Create frames
    prev_neurons = []
    for step in track(
        np.arange(N_FRAMES), total=N_FRAMES, description="Generating frames..."
    ):
        if step % N_frames_for_change == 0:  # change neurons every N framse

            # reset neurons from previous set of neurons
            for neuron in prev_neurons:
                for component, actor in neuron.items():
                    actor.alpha(minalpha)
                    actor.color(darkcolor)
            prev_neurons = []

            # highlight new neurons
            neurons = choices(neurons_actors, k=N_neurons_in_frame)
            for n, neuron in enumerate(neurons):
                color = colorMap(
                    n, "Greens_r", vmin=-2, vmax=N_neurons_in_frame + 3
                )
                for component, actor in neuron.items():
                    actor.alpha(1)
                    actor.color(color)
                prev_neurons.append(neuron)

        # Move scene camera between 3 cameras
        scene.plotter.moveCamera(cam1, cam2, frac[step])
        if frac[step] == 1:
            cam1 = cam3

        # Update rendered window
        time.sleep(0.1)
        scene.render(zoom=zoom[step], interactive=False, video=True)
    scene.close()


@pytest.mark.slow
def test_video():
    from brainrender.animation.video import BasicVideoMaker as VideoMaker

    scene = Scene()

    # Create an instance of VideoMaker with our scene
    vm = VideoMaker(scene, niters=10)

    # Make a video!
    vm.make_video(
        elevation=1, roll=5
    )  # specify how the scene rotates at each frame


@pytest.mark.slow
def test_custom_video():
    from brainrender.animation.video import CustomVideoMaker

    # --------------------------------- Variables -------------------------------- #
    N_FRAMES = 20

    # Variables to specify camera position at each frame
    zoom = np.linspace(1, 1.35, N_FRAMES)
    frac = np.zeros_like(
        zoom
    )  # for camera transition, interpolation value between cameras
    frac[:10] = np.linspace(0, 1, 10)
    frac[10:] = np.linspace(1, 0, len(frac[10:]))

    # ------------------------------- Create scene ------------------------------- #
    scene = Scene(display_inset=True, use_default_key_bindings=True)

    filepaths, data = scene.atlas.download_streamlines_for_region("TH")
    scene.add_brain_regions(["TH"], alpha=0.2)

    # Create new cameras
    cam1 = buildcam(sagittal_camera)
    cam2 = buildcam(top_camera)
    cam3 = buildcam(
        dict(
            position=[1862.135, -4020.792, -36292.348],
            focal=[6587.835, 3849.085, 5688.164],
            viewup=[0.185, -0.97, 0.161],
            distance=42972.44,
            clipping=[29629.503, 59872.10],
        )
    )

    # Iniziale camera position
    scene.plotter.moveCamera(cam1, cam2, frac[0])

    # ------------------------------- Create frames ------------------------------ #
    def frame_maker(scene=None, video=None, videomaker=None):
        for step in track(
            np.arange(N_FRAMES),
            total=N_FRAMES,
            description="Generating frames...",
        ):
            # Move scene camera between 3 cameras
            if step < 150:
                scene.plotter.moveCamera(cam1, cam2, frac[step])
            else:
                scene.plotter.moveCamera(cam3, cam2, frac[step])

            # Add frame to video
            scene.render(zoom=zoom[step], interactive=False, video=True)
            video.addFrame()
        return video

    # ---------------------------------------------------------------------------- #
    #                                  Video maker                                 #
    # ---------------------------------------------------------------------------- #
    vm = CustomVideoMaker(scene, save_name="streamlines_animation")
    vm.make_video(frame_maker)



import os
import numpy as np
import pytest
from brainrender.scene import Scene

from brainrender.Utils.camera import get_camera_params, set_camera
from brainrender.Utils.data_io import listdir, get_subdirs
from brainrender.Utils.data_manipulation import get_slice_coord
from brainrender.colors import (
    get_random_colormap,
    get_n_shades_of,
    getColor,
    getColorName,
    colorMap,
    check_colors,
)
from brainrender.Utils.actors_funcs import edit_actor
from brainrender.Utils.ruler import ruler


@pytest.fixture
def scene():
    return Scene()


def test_ruler(scene):
    # add brain regions
    mos, hy = scene.add_brain_regions(["MOs", "HY"], alpha=0.2)
    mos.wireframe()
    hy.wireframe()

    # Get center of mass of the two regions
    p1 = scene.atlas.get_region_CenterOfMass("MOs")
    p2 = scene.atlas.get_region_CenterOfMass("HY")

    # Use the ruler class to display the distance between the two points
    """
        Brainrender units are in micrometers. To display the distance
        measure instead we will divide by a factor of 1000 using 
        the unit_scale argument.
    """

    # Add a ruler form the brain surface
    scene.add_ruler_from_surface(p2)

    # Add a ruler between the two regions
    scene.add_actor(ruler(p1, p2, unit_scale=0.01, units="mm"))


def test_get_camera_params(scene):
    if not isinstance(get_camera_params(scene), dict):
        raise ValueError

    if not isinstance(get_camera_params(camera=scene.plotter.camera), dict):
        raise ValueError

    camera_dict = get_camera_params(scene)

    try:
        set_camera(scene, camera_dict)
    except ValueError as e:
        raise ValueError(
            f"Failed to produce a camera params dict that can be used to set camera:\n {e}"
        )


def test_io():
    fld = os.getcwd()
    if not isinstance(listdir(fld), list):
        raise ValueError(f"listdir returned: {listdir(fld)}")
    if not isinstance(get_subdirs(fld), list):
        raise ValueError(f"get_subdirs returned: {get_subdirs(fld)}")


def test_get_slice():
    get_slice_coord([0.0, 100.0], 0.5)


def test_colors():
    if not isinstance(get_random_colormap(), str):
        raise ValueError

    cols = get_n_shades_of("green", 40)
    if not isinstance(cols, list):
        raise ValueError
    if len(cols) != 40:
        raise ValueError

    getColor("orange")
    getColor([1, 1, 1])
    getColor("k")
    getColor("#ffffff")
    getColor(7)
    getColor(-7)

    getColorName("#ffffff")

    cols = colorMap([0, 1, 2])
    if not isinstance(cols, (list, np.ndarray)):
        raise ValueError
    if len(cols) != 3:
        raise ValueError

    c = colorMap(3, vmin=-3, vmax=4)

    check_colors(cols)
    check_colors(c)


def test_edit_actors(scene):
    actor = scene.add_brain_regions("TH")

    edit_actor(
        actor,
        wireframe=True,
        color="red",
        line=True,
        line_kwargs={"lw": 2},
        upsample=True,
        downsample=False,
    )
    edit_actor(
        actor, solid=True, downsample=True, smooth=True,
    )



"""
    This example shows how to cut actors in the scene using a plane
    oriented along the sagittal axis
"""

import brainrender

brainrender.SHADER_STYLE = "cartoon"

from brainrender.scene import Scene


scene = Scene(title="cut with plane")

# Add some actors
th = scene.add_brain_regions(["STR", "TH"], alpha=0.5)

# Cut with plane
scene.cut_actors_with_plane(
    "sagittal", showplane=False
)  # Set showplane to True if you want to see the plane location

# Add a silhouette around each actor to emphasize the cut location
scene.add_silhouette(*th, lw=3)


scene.render(camera="top")



""" 
    This tutorial shows how to show the position of labelled cells in brainrender
    and color them based on the region they are in. 
"""
import pandas as pd

import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene
from brainrender.Utils.scene_utils import get_n_random_points_in_region

# Create a scene
scene = Scene(
    title="labelled cells"
)  # specify that you want a view from the top


# Gerate the coordinates of N cells across 3 regions
_regions = ["MOs", "VISp", "ZI"]
N = 1000  # getting 1k cells per region, but brainrender can deal with >1M cells easily.

# Get fake cell coordinates
cells, regions = [], []  # to store x,y,z coordinates
for region in _regions:
    region_cells = get_n_random_points_in_region(
        scene.atlas, region=region, N=N
    )
    cells.extend(region_cells)
    regions.extend([region for i in region_cells])
x, y, z = (
    [c[0] for c in cells],
    [c[1] for c in cells],
    [c[2] for c in cells],
)
cells = pd.DataFrame(
    dict(x=x, y=y, z=z, region=regions)
)  # ! <- coordinates should be stored as a pandas dataframe

# Add cells

scene.add_cells(cells, color_by_region=True, radius=50, res=25)

# render
scene.render()



""" 
    This tutorial shows how to create and render a brainrender scene with some brain regions
"""
import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

# Create a scene
scene = Scene(title="brain regions")

# Add the whole thalamus in gray
scene.add_brain_regions(["TH"], alpha=0.15)

# Add VAL nucleus in wireframe style with the allen color
val = scene.add_brain_regions(["VAL"], use_original_color=True)
val.wireframe()

scene.render()



import pytest
import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.morphology.visualise import MorphologyScene


@pytest.fixture
def scene():
    return MorphologyScene(title="A neuron")


def test_color_options(scene):
    scene.add_neurons(
        "Examples/example_files/neuron4.swc",
        color=dict(soma="red", dendrites="orangered", axon=[0.4, 0.4, 0.4],),
    )

    scene.add_neurons(
        "Examples/example_files/neuron4.swc",
        color=[dict(soma="red", dendrites="orangered", axon=[0.4, 0.4, 0.4],)],
    )

    scene.add_neurons("Examples/example_files/neuron4.swc",)

    scene.add_neurons("Examples/example_files/neuron4.swc", color="salmon")

    scene.add_neurons("Examples/example_files/neuron4.swc", color="Reds")

    scene.add_neurons("Examples/example_files/neuron4.swc", color=["seagreen"])


def test_render_options(scene):
    scene.add_neurons(
        "Examples/example_files/neuron4.swc", neurite_radius=None,
    )

    scene.add_neurons(
        ["Examples/example_files/neuron4.swc"], neurite_radius=None,
    )

    scene.add_neurons(
        "Examples/example_files/neuron4.swc", neurite_radius=None,
    )

    scene.add_neurons(
        "Examples/example_files/neuron4.swc",
        display_axon=False,
        neurite_radius=None,
    )

    scene.add_neurons(
        "Examples/example_files/neuron4.swc",
        display_dendrites=False,
        neurite_radius=None,
    )

    scene.add_neurons(
        "Examples/example_files/neuron4.swc",
        display_axon=False,
        display_dendrites=False,
        neurite_radius=None,
    )



from accepts import accepts

# ---------------------------------------------------------------------------- #
#                                PYTHON OBJECTS                                #
# ---------------------------------------------------------------------------- #


@accepts((list, tuple))
def flatten(lst):
    """
        Flattens a nested list
    """
    return [item for sublist in lst for item in sublist]


@accepts((list, tuple))
def return_list_smart(lst):
    """
        If the list has length > returns the list
        if it has length == 1 it returns the element
        if it has length == 0 it returns None
    """
    if len(lst) > 1:
        return lst
    elif len(lst) == 1:
        return lst[0]
    else:
        return None


@accepts(dict)
def return_dict_smart(dct):
    keys = list(dct.keys())
    if len(keys) == 1:
        return dct[keys[0]]
    elif len(keys) == 0:
        return None
    else:
        return dct


@accepts((list, dict), (list, dict))
def is_any_item_in_list(L1, L2):
    """
    Checks if an item in a list is in another  list

    :param L1: 
    :param L2: 

    """
    # checks if any item of L1 is also in L2 and returns false otherwise
    inboth = [i for i in L1 if i in L2]
    if inboth:
        return True
    else:
        return False


# ---------------------------------------------------------------------------- #
#                             BRAINRENDER SPECIFIC                             #
# ---------------------------------------------------------------------------- #


def get_slice_coord(bounds, n):
    """
    Given the bounds of an actor, return the point that
    corresponds to the n% of the bounds range
    

    :param bounds: should be a list of two floats
    :param n: n should be a float in range 0, 1

    """
    if (
        not isinstance(bounds, (list, tuple))
        or not isinstance(bounds[0], float)
        or not isinstance(bounds[1], float)
    ):
        raise ValueError(
            "bounds should be a list or tuple of floats: {}".format(bounds)
        )
    if not isinstance(n, (int, float)):
        raise ValueError("n should be a float")
    if n < 0 or n > 1:
        raise ValueError("n should be in range [0, 1]")

    b0, b1 = bounds
    delta = b1 - b0

    return b0 + delta * n


def make_optic_canula_cylinder(
    atlas,
    root,
    target_region=None,
    pos=None,
    offsets=(0, 0, -500),
    hemisphere="right",
    color="powderblue",
    radius=350,
    alpha=0.5,
    **kwargs,
):

    """
        Creates a cylindrical vedo actor to scene to render optic cannulas. By default
        this is a semi-transparent blue cylinder centered on the center of mass of
        a specified target region and oriented vertically.

        :param target_region: str, acronym of target region to extract coordinates
            of implanted fiber. By defualt the fiber will be centered on the center
            of mass of the target region but the offset arguments can be used to
            fine tune the position. Alternative pass a 'pos' argument with AP-DV-ML coords.
        :param pos: list or tuple or np.array with X,Y,Z coordinates. Must have length = 3.
        :param x_offset, y_offset, z_offset: int, used to fine tune the coordinates of 
            the implanted cannula.
        :param **kwargs: used to specify which hemisphere the cannula is and parameters
            of the rendered cylinder: color, alpha, rotation axis...
    """

    # Get coordinates of brain-side face of optic cannula
    if target_region is not None:
        pos = atlas.get_region_CenterOfMass(
            target_region, unilateral=True, hemisphere=hemisphere
        )
    elif pos is None:
        print(
            "No 'pos' or 'target_region' arguments were \
                        passed to 'add_optic_cannula', nothing to render"
        )
        return

    # Offset position
    for i, offset in enumerate(offsets):
        pos[i] += offset

    # Get coordinates of upper face
    bounds = root.bounds()
    top = pos.copy()
    top[1] = bounds[2] - 500

    # Create actor
    return dict(pos=[top, pos], c=color, r=radius, alpha=alpha, **kwargs)


def get_cells_in_region(atlas, cells, region):
    """
        Selects the cells that are in a list of user provided 
        brain regions from a dataframe of cell locations

        :param cells: pd.DataFrame of cells x,y,z coordinates
    """
    if isinstance(region, list):
        region_list = []
        for reg in region:
            region_list.extend(list(atlas.get_structure_descendants(reg)))
    else:
        region_list = list(atlas.get_structure_descendants(region))
    return cells[cells.region.isin(region_list)]



"""
    This example shows how to measure and display
    the distance between two points.
"""


from brainrender.scene import Scene
import brainrender
from brainrender import ruler


brainrender.SHOW_AXES = True

# Create Scene
scene = Scene()

# add brain regions
mos, hy = scene.add_brain_regions(["MOs", "HY"], alpha=0.2)
mos.wireframe()
hy.wireframe()

# Get center of mass of the two regions
p1 = scene.atlas.get_region_CenterOfMass("MOs")
p2 = scene.atlas.get_region_CenterOfMass("HY")


# Use the ruler class to display the distance between the two points
"""
    Brainrender units are in micrometers. To display the distance
    measure instead we will divide by a factor of 1000 using 
    the unit_scale argument.
"""

# Add a ruler form the brain surface
scene.add_ruler_from_surface(p2)

# Add a ruler between the two regions
scene.add_actor(ruler(p1, p2, unit_scale=0.01, units="mm"))

# render
scene.render()



import pytest

try:
    from brainrender.ABA.atlas_images import ImageDownload
except ImportError:
    ImageDownload = None
from tempfile import TemporaryDirectory


@pytest.fixture
def imd():
    if ImageDownload is not None:
        return ImageDownload()


def test_imd(imd):
    if imd is not None:
        imd.get_atlas_by_name(imd.mouse_coronal)
        imd.get_products_by_species("Mouse")


def test_get_experiments(imd):
    if imd is not None:
        imd.get_experimentsid_by_productid(26)


def test_get_atlas_images(imd):
    if imd is not None:
        imd.get_atlasimages_by_atlasid(imd.mouse_coronal_atlas_id)
        imd.get_atlasimages_by_atlasid(imd.mouse_sagittal_atlas_id)


@pytest.mark.slow
def test_download_imgs(imd):
    if imd is not None:
        tmp_dir = TemporaryDirectory()
        imd.download_images_by_atlasid(
            tmp_dir.name, imd.mouse_coronal_atlas_id, debug=True
        )
        imd.download_images_by_atlasid(
            tmp_dir.name,
            imd.mouse_sagittal_atlas_id,
            debug=True,
            downsample=4,
            snames=True,
            annotated=False,
            atlas_svg=False,
        )
        tmp_dir.cleanup()



try:
    import cv2
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "You need opencv to save videos in brainrender, please install opencv with: "
        + "pip install opencv-python"
    )

"""
    Collection of utility functions useful when deaing with cv2
    'cap' objects
"""


def cap_set_frame(cap, frame_number):
    """
        Sets an opencv video capture object to a specific frame
    """
    cap.set(1, frame_number)


def get_cap_selected_frame(cap, show_frame):
    """ 
        Gets a frame from an opencv video capture object to a specific frame
    """
    cap_set_frame(cap, show_frame)
    ret, frame = cap.read()

    if not ret:
        return None
    else:
        return frame


def get_video_params(cap):
    """ 
        Gets video parameters from an opencv video capture object
    """
    if isinstance(cap, str):
        cap = cv2.VideoCapture(cap)

    frame = get_cap_selected_frame(cap, 0)
    if frame.shape[1] == 3:
        is_color = True
    else:
        is_color = False
    cap_set_frame(cap, 0)

    nframes = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fps = cap.get(cv2.CAP_PROP_FPS)
    return nframes, width, height, fps, is_color


def open_cvwriter(
    filepath, w=None, h=None, framerate=None, format=".mp4", iscolor=False
):
    """
        Creats an instance of cv.VideoWriter to write frames to video using python opencv
        :param filepath: str, path to file to save
        :param w,h: width and height of frame in pixels
        :param framerate: fps of output video
        :param format: video format
        :param iscolor: bool, set as true if images are rgb, else false if they are gray
    """
    try:
        if "avi" in format:
            fourcc = cv2.VideoWriter_fourcc("M", "J", "P", "G")
        else:
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        videowriter = cv2.VideoWriter(
            filepath, fourcc, framerate, (w, h), iscolor
        )
    except:
        raise ValueError("Could not create videowriter")
    else:
        return videowriter


def save_videocap_to_video(cap, savepath, fmt, fps=30, iscolor=True):
    """
        Saves the content of a videocapture opencv object to a file
    """
    if "." not in fmt:
        fmt = "." + fmt
    # Creat video writer
    nframes, width, height, _, _ = get_video_params(cap)
    writer = open_cvwriter(
        savepath, w=width, h=height, framerate=fps, format=fmt, iscolor=iscolor
    )

    # Save frames
    while True:
        ret, frame = cap.read()
        if not ret:
            break

        if frame.shape[2] == 4:  # rgba to rgb
            frame = cv2.cvtColor(frame, cv2.COLOR_RGBA2RGB)

        writer.write(frame)

    # Release everything if job is finished
    cap.release()
    writer.release()



import numpy as np
from vedo import Mesh
from accepts import accepts

""" 
    Collection of functions to edit actors looks and other features.
"""


@accepts(Mesh)
def get_actor_bounds(actor):
    """
        Gets the AP-DV-ML bounds of an actor
    """
    bounds = actor.bounds()
    return [
        [bounds[0], bounds[1]],
        [bounds[2], bounds[3]],
        [bounds[4], bounds[5]],
    ]


@accepts(Mesh)
def get_actor_midpoint(actor):
    """ 
        Get's the coordinates of the midpoint of an
        actor (vedo.Mesh)
    """
    return [np.mean(bounds) for bounds in get_actor_bounds(actor)]


# TODO mirror around point would mean
@accepts(Mesh, (list, np.ndarray, float), str)
def mirror_actor_at_point(actor, point, axis="x"):
    """
    Mirror an actor around a point

    :param actor: 
    :param point: 
    :param axis:  (Default value = 'x')

    """
    if not isinstance(actor, dict):
        coords = actor.points()
        if axis == "x":
            shifted_coords = [
                [c[0], c[1], point + (point - c[2])] for c in coords
            ]
        elif axis == "y":
            shifted_coords = [
                [c[0], point + (point - c[1]), c[2]] for c in coords
            ]
        elif axis == "z":
            shifted_coords = [
                [point + (point - c[0]), c[1], c[2]] for c in coords
            ]

        actor.points(shifted_coords)
        actor = actor.mirror(
            axis="n"
        )  # to make sure that the mirrored actor looks correctly
        return actor
    else:
        mirrored_actor = {}
        for n, a in actor.items():
            coords = a.points()
            if axis == "x":
                shifted_coords = [
                    [c[0], c[1], point + (point - c[2])] for c in coords
                ]
            elif axis == "y":
                shifted_coords = [
                    [c[0], point + (point - c[1]), c[2]] for c in coords
                ]
            elif axis == "z":
                shifted_coords = [
                    [point + (point - c[0]), c[1], c[2]] for c in coords
                ]

            a.points(shifted_coords)
            a = a.mirror(
                axis="n"
            )  # to make sure that the mirrored actor looks correctly
            mirrored_actor[n] = actor
        return mirrored_actor


@accepts(Mesh, (int, float), (type(None), str))
def set_line(actor, lw=None, c=None):
    """
    set an actor's look to specify the line width and color

    :param actor: 
    :param lw:  (Default value = None)
    :param c:  (Default value = None)

    """
    if lw is not None:
        actor.lw(lineWidth=lw)
    if c is not None:
        actor.lc(lineColor=c)


def edit_actor(
    actor,
    wireframe=False,
    solid=False,
    color=False,
    line=False,
    line_kwargs={},
    upsample=False,
    downsample=False,
    smooth=False,
):
    """
    Apply a set of functions to edit an actor's look. 

    :param actor: 
    :param wireframe: if true, change look to wireframe (Default value = False)
    :param solid: if true change look to soi=lid (Default value = False)
    :param color: specify new color (Default value = False)
    :param line: if true, edit the line's look (Default value = False)
    :param line_kwargs: specify width and color of line (Default value = {})
    :param upsample: if true, increase resolution (Default value = False)
    :param downsample: if true, decrease resolution (Default value = False)
    :param smooth: if true, smoothen actor (Default value = False)

    """

    if wireframe:
        actor.wireframe(value=True)
    if solid:
        actor.wireframe(value=False)
    if color:
        actor.color(c=color)
    if line:
        set_line(actor, **line_kwargs)
    if upsample:
        actor.subdivide(N=1)
    if downsample:
        actor.decimate(fraction=0.5)
    if smooth:
        actor.smoothLaplacian(niter=15)



from brainrender.scene import Scene
from brainrender.colors import makePalette

from morphapi.api.mpin_celldb import MpinMorphologyAPI


"""
    This examples shows how to download and render neurons
    reconstructions aligned to the zebra fish atlas.
"""

# Use morphapi's API to download the neurons
api = MpinMorphologyAPI()

# the first time you use this dataset you'll have to download the data
# api.download_dataset()

# ----------------------------- Download dataset ----------------------------- #
"""
    If it's the first time using this API, you'll have to download the dataset
    with all of the neurons' data.
"""
# Select a few neurons for a specific brain region
neurons_ids = api.get_neurons_by_structure(837)
neurons = api.load_neurons(neurons_ids)

# Create meshes for each neuron

neurons = [
    neuron.create_mesh(soma_radius=1, neurite_radius=1)[1]
    for neuron in neurons
]
colors = makePalette(len(neurons), "salmon", "powderblue")

# ------------------------------- Visualisation ------------------------------ #
scene = Scene(atlas="mpin_zfish_1um", add_root=True, camera="sagittal2")
scene.add_neurons(neurons, color=colors)
scene.render()



"""
    This tutorial shows you how to render efferent mesoscale connectivity
    data from the Allen mouse connectome project as streamlines. 

"""

from brainrender.scene import Scene

# Start by creating a scene with the allen brain atlas atlas
scene = Scene(title="streamlines")


# Download streamlines data for injections in the CA1 field of the hippocampus
filepaths, data = scene.atlas.download_streamlines_for_region("CA1")


scene.add_brain_regions(["CA1"], use_original_color=True, alpha=0.2)

# you can pass either the filepaths or the data
scene.add_streamlines(data, color="darkseagreen", show_injection_site=False)

scene.render(camera="sagittal", zoom=1)



"""
    This tutorial shows how to use flags (strings of text that appear when you mose hoover over an actor)
    to add more information about the actors in your rendering. 
"""


import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

# Create a scene
scene = Scene(title="labels")

# add_brain_regions can be used to add labels directly
scene.add_brain_regions("VAL", add_labels=True)

# you can also use scene.add_actor_label
mos = scene.add_brain_regions("MOs")

# Add another label, this time make it gray and shift it slightly
scene.add_actor_label(mos, "MOs", size=400, color="blackboard", xoffset=250)

scene.render()



""" 
    This tutorial shows how to add `vedo` actors to your scene. 
"""
import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

# Create a scene
scene = Scene(title="Adding actors")

# There's two ways to add an actor to a scene.
# Using the '+' operator with a filepath or an instance of `vedo.Mesh`
scene + "Examples/example_files/root.obj"  # or scene += some_actor

# or using the scene class `add_actor` method
scene.add_actor(scene.root.clone().color("r"))


# You can check how many actors there are in a scene with
print(scene)  # or len(scene)

# and you can acces all actors in a scene with:
for actor in scene:
    print(f"this is an actor!: {actor}")


scene.render()
